import abc
from typing import Type, TypeVar, Dict

from pydantic import BaseModel

from pki_tools.exceptions import (
    MissingInit,
)

from loguru import logger

CryptoObject = TypeVar("CryptoObject")


class CryptoParser(BaseModel, abc.ABC):
    """
    CryptoParser is an abstract class used by all the types
    parsing cryptography objects into [pki_tools][pki_tools.types.certificate]
    pydantic classes.
    """

    _x509_obj: CryptoObject

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if "_x509_obj" not in kwargs:
            try:
                self._x509_obj = self._to_cryptography()
            except MissingInit:
                logger.trace("Can't create crypto object before init")

    @classmethod
    @abc.abstractmethod
    def from_cryptography(
        cls: Type["CryptoParser"], crypto_obj: CryptoObject
    ) -> "CryptoParser":
        """
        Parses a cryptography x509 object into a
        [CryptoParser][pki_tools.types.crypto_parser.CryptoParser]

        Arguments:
             crypto_obj: The cryptography object

        Returns:
            CryptoParser object
        """

    @abc.abstractmethod
    def _to_cryptography(self) -> CryptoObject:
        """
        Creates a x509 cryptography object from this class

        Returns: A x509 cryptography object
        """

    @abc.abstractmethod
    def _string_dict(self) -> Dict[str, str]:
        """
        Creates a dict representation of the object

        Returns: A dict containing all the keys in the CryptoParser
        """


class InitCryptoParser(CryptoParser, abc.ABC):
    """
    Extends the CryptoParser into an object that requires initialization
    before it can be used (while created as a
    [pki_tools][pki_tools.types.certificate] object and not loaded from
    cryptography). This can, for example, be a Certificate that needs
    to be signed with a KeyPair containing the private key.

    Attempt to e.g. dumping a certificate to a PEM string without using the
    sign (init) function first will result in a
    [MissingInit][pki_tools.exceptions.MissingInit] exception.
    """

    _init_func: str = "sign"

    @property
    def _crypto_object(self) -> CryptoObject:
        if not hasattr(self, "_x509_obj") or self._x509_obj is None:
            init_func = f"{self.__class__.__name__}.{self._init_func}"
            raise MissingInit(f"Please use the {init_func} first")

        return self._x509_obj



import time
from functools import lru_cache

from loguru import logger

from pki_tools.types.chain import Chain
from pki_tools.types.certificate import Certificate
from pki_tools.types.signature_algorithm import HashAlgorithm
from pki_tools.types.utils import HTTPX_CLIENT
from pki_tools.exceptions import (
    ExtensionMissing,
    OcspInvalidResponseStatus,
    OcspFetchFailure,
    Error,
)
from pki_tools.types.extensions import (
    AuthorityInformationAccess,
    UniformResourceIdentifier,
    AccessDescriptionId,
)
from pki_tools.types.ocsp import OCSPResponse, OCSPRequest

OCSP_ALGORITHMS_TO_CHECK = [
    HashAlgorithm(name="SHA256"),
    HashAlgorithm(name="SHA1"),
    HashAlgorithm(name="SHA512"),
    HashAlgorithm(name="SHA224"),
    HashAlgorithm(name="SHA384"),
]


def _is_revoked_multiple_issuers(
    cert: Certificate,
    cert_issuer: Chain,
    ocsp_issuer: Chain,
    ocsp_res_cache_seconds: int = 3600,
):
    cert_issuer.check_chain()
    ocsp_issuer.check_chain()

    issuer = cert_issuer.get_issuer(cert)

    log = logger.bind(
        cert=cert.pem_string,
        serial=cert.serial_number,
    )

    if cert.extensions.authority_information_access is None:
        log.debug("OCSP extension missing")
        raise ExtensionMissing()

    for i, alg in enumerate(OCSP_ALGORITHMS_TO_CHECK):
        try:
            req_path = _construct_req_path(cert, issuer, alg)

            return _check_ocsp_status(
                cert.extensions.authority_information_access,
                req_path,
                cert,
                ocsp_issuer,
                ocsp_res_cache_seconds=ocsp_res_cache_seconds,
            )
        except OcspInvalidResponseStatus:
            log.bind(alg=alg.name.value).debug(
                "OCSP check failed, trying another algorithm"
            )
            if i + 1 == len(OCSP_ALGORITHMS_TO_CHECK):
                log.bind(
                    algs=[alg.name.value for alg in OCSP_ALGORITHMS_TO_CHECK]
                ).debug("All algorithms check failed")
                raise

    return False


def _construct_req_path(cert, issuer_cert, alg):
    req = OCSPRequest(hash_algorithm=alg)
    req.create(cert, issuer_cert)
    return req.request_path


def _check_ocsp_status(
    aia: AuthorityInformationAccess,
    req_path,
    cert: Certificate,
    issuer_chain: Chain,
    ocsp_res_cache_seconds: int = 3600,
):
    log = logger.bind(serial=cert.hex_serial)

    checked_status = False
    for access_description in aia:
        if access_description.access_method != AccessDescriptionId.OCSP:
            logger.trace(
                "Access method is not OCSP, "
                "try checking next access description"
            )
            continue

        if not isinstance(
            access_description.access_location, UniformResourceIdentifier
        ):
            continue

        checked_status = True

        server = access_description.access_location.value

        cache_ttl = round(time.time() / ocsp_res_cache_seconds)
        ocsp_res = _get_ocsp_status(
            f"{server}/{req_path}", cache_ttl=cache_ttl
        )

        _verify_ocsp_signature(ocsp_res, issuer_chain)

        if ocsp_res.is_revoked:
            log.bind(
                date=str(ocsp_res._x509_obj.revocation_time),
            ).debug("Certificate revoked")
            return True
        else:
            return False

    if not checked_status:
        raise ExtensionMissing()

    log.debug("Certificate valid")
    return False


@lru_cache(maxsize=None)
def _get_ocsp_status(uri, cache_ttl=None) -> OCSPResponse:
    ret = HTTPX_CLIENT.get(
        uri, headers={"Content-Type": "application/ocsp-request"}
    )

    log = logger.bind(status=ret.status_code)
    if ret.status_code != 200:
        log.error("OCSP status fetch failed")
        raise OcspFetchFailure(
            f"Unexpected response status code: {ret.status_code}"
        )

    ocsp_res = OCSPResponse.from_der_bytes(ret.content)

    if not ocsp_res.is_successful:
        log.bind(res=ocsp_res.response_status).debug("Invalid OCSP response")
        raise OcspInvalidResponseStatus(
            f"Invalid OCSP Response status: {ocsp_res.response_status}"
        )

    return ocsp_res


def _verify_ocsp_signature(ocsp_response: OCSPResponse, issuer_chain: Chain):
    for issuer_cert in issuer_chain.certificates:
        cert_public_hash = ocsp_response.hash_with_alg(
            issuer_cert.der_public_key
        )

        if cert_public_hash == ocsp_response.issuer_key_hash:
            break
    else:
        logger.error("Couldn't find OCSP response issuer")
        raise Error("Couldn't find OCSP response issuer")

    issuer_cert.verify_signature(ocsp_response)



import importlib
import typing
from enum import Enum
from typing import List, Optional, Iterable, Union, Type, Dict

from cryptography import x509
from cryptography.hazmat._oid import (
    ExtensionOID,
    AuthorityInformationAccessOID,
)
from cryptography.hazmat.bindings._rust import ObjectIdentifier

from cryptography.x509.extensions import (
    ExtensionNotFound,
    ExtensionTypeVar,
)

from loguru import logger
from pydantic import Field, ConfigDict

from pki_tools.types.crypto_parser import CryptoParser
from pki_tools.types.name import Name
from pki_tools.types.utils import _byte_to_hex, _hex_to_byte

GENERAL_NAME_MODULE = importlib.import_module("cryptography.x509.general_name")
EXTENSIONS_MODULE = importlib.import_module("cryptography.x509.extensions")


class Extension(CryptoParser):
    """
    Represents a cryptographic extension.

    Attributes:
        critical: Indicates whether the extension is critical.
    """

    critical: Optional[bool] = False

    @property
    def name(self) -> str:
        """
        Get the name of the extension.

        Returns:
           str: The name of the extension.
        """
        name = "".join(
            [" " + c if c.isupper() else c for c in self.__class__.__name__]
        )[1:]
        if self.critical:
            name += " (critical)"
        return name


class GeneralName(CryptoParser):
    """
    Represents a general name.

    Attributes:
        name: The name of the general name.
        value: The value of the general name.
    """

    name: str
    value: Union[str, Name]

    @classmethod
    def from_cryptography(
        cls: Type["GeneralName"], crypto_obj: x509.GeneralName
    ) -> "GeneralName":
        """
        Create a GeneralName instance from a cryptography GeneralName object.

        Args:
            crypto_obj: The cryptography GeneralName object.

        Returns:
            The GeneralName instance.
        """
        if isinstance(crypto_obj, x509.DNSName):
            return DnsName(crypto_obj.value)
        elif isinstance(crypto_obj, x509.IPAddress):
            return IpAddress(str(crypto_obj.value))
        elif isinstance(crypto_obj, x509.DirectoryName):
            return DirectoryName.from_cryptography(crypto_obj)
        elif isinstance(crypto_obj, x509.OtherName):
            return OtherName.from_cryptography(crypto_obj)
        elif isinstance(crypto_obj, x509.RegisteredID):
            return RegisteredId.from_cryptography(crypto_obj)

        return globals()[crypto_obj.__class__.__name__](crypto_obj.value)

    def _to_cryptography(self) -> x509.GeneralName:
        return getattr(GENERAL_NAME_MODULE, self.name)(self.value)

    def _string_dict(self) -> typing.Dict[str, str]:
        return {
            "name": self.name,
            "value": self.value,
        }


class DnsName(GeneralName):
    """
    Represents a DNS name.

    Attributes:
        value: The DNS name value.
    """

    def __init__(self, value: str):
        super().__init__(name="DNSName", value=value)


class DirectoryName(GeneralName):
    """
    Represents a directory name.

    Attributes:
        value: The directory name value.
    """

    def __init__(self, value: Name):
        super().__init__(name="DirectoryName", value=value)

    @classmethod
    def from_cryptography(
        cls: Type["DirectoryName"], crypto_obj: x509.GeneralName
    ) -> "DirectoryName":
        """
        Create a DirectoryName instance from a cryptography DirectoryName
            object.

        Args:
            crypto_obj: The cryptography DirectoryName object.

        Returns:
            The DirectoryName instance.
        """
        return cls(value=Name.from_cryptography(crypto_obj.value))

    def _to_cryptography(self) -> x509.GeneralName:
        return x509.DirectoryName(self.value._to_cryptography())


class IpAddress(GeneralName):
    """
    Represents an IP address. Can either be a IPv4/IPv6 single address or
    network.

    Attributes:
        value: The IP address or network
    """

    def __init__(self, value: str):
        super().__init__(name="IPAddress", value=value)

    def _to_cryptography(self) -> x509.GeneralName:
        cls_name = "IPv4"
        if ":" in self.value:
            cls_name = "IPv6"

        if "/" in self.value:
            cls_name += "Network"
        else:
            cls_name += "Address"

        module = importlib.import_module("ipaddress")
        value = getattr(module, cls_name)(self.value)
        return x509.IPAddress(value)


class OtherName(GeneralName):
    """
    Represents an other name.

    Attributes:
        value: The other name value.
        oid: The object identifier (OID) of the other name.
    """

    oid: str

    def __init__(self, value: str, oid: str):
        super().__init__(name="OtherName", value=value, oid=oid)

    @classmethod
    def from_cryptography(
        cls: Type["OtherName"], crypto_obj: x509.OtherName
    ) -> "OtherName":
        """
        Create a OtherName instance from a cryptography OtherName object.

        Args:
            crypto_obj: The cryptography OtherName object.

        Returns:
            The OtherName instance.
        """
        return cls(
            value=_byte_to_hex(crypto_obj.value),
            oid=crypto_obj.type_id.dotted_string,
        )

    def _to_cryptography(self) -> x509.OtherName:
        return x509.OtherName(
            type_id=x509.ObjectIdentifier(self.oid),
            value=_hex_to_byte(self.value),
        )


class RFC822Name(GeneralName):
    """
    Represents an RFC822 name.

    Attributes:
        value: The RFC822 name value.
    """

    def __init__(self, value: str):
        super().__init__(name="RFC822Name", value=value)


class RegisteredId(GeneralName):
    """
    Represents a registered ID.

    Attributes:
        value: The registered ID value.
    """

    def __init__(self, value: str):
        super().__init__(name="RegisteredID", value=value)

    @classmethod
    def from_cryptography(
        cls: Type["RegisteredId"], crypto_obj: x509.RegisteredID
    ) -> "RegisteredId":
        """
        Create a RegisteredId instance from a cryptography RegisteredId object.

        Args:
            crypto_obj: The cryptography RegisteredId object.

        Returns:
            The RegisteredId instance.
        """
        return cls(value=str(crypto_obj.value.dotted_string))

    def _to_cryptography(self) -> x509.RegisteredID:
        return x509.RegisteredID(x509.ObjectIdentifier(self.value))


class UniformResourceIdentifier(GeneralName):
    """
    Represents a uniform resource identifier (URI).

    Attributes:
        value: The URI value.
    """

    def __init__(self, value: str):
        super().__init__(name="UniformResourceIdentifier", value=value)


class AuthorityKeyIdentifier(Extension):
    """
    Represents an authority key identifier extension.

    Attributes:
        key_identifier: The key identifier.
        authority_cert_issuer: The issuer of the authority certificate.
        authority_cert_serial_number: The serial number of the authority
            certificate.
    """

    key_identifier: Optional[bytes]
    authority_cert_issuer: Optional[List[GeneralName]]
    authority_cert_serial_number: Optional[int]

    @classmethod
    def from_cryptography(
        cls: Type["AuthorityKeyIdentifier"],
        extension: x509.AuthorityKeyIdentifier,
    ) -> "AuthorityKeyIdentifier":
        """
        Create an AuthorityKeyIdentifier instance from a cryptography
        AuthorityKeyIdentifier object.

        Args:
            extension: The cryptography AuthorityKeyIdentifier object.

        Returns:
            The AuthorityKeyIdentifier instance.
        """
        issuers = None
        if extension.authority_cert_issuer is not None:
            issuers = []
            for general_name in extension.authority_cert_issuer:
                issuers.append(GeneralName.from_cryptography(general_name))

        return cls(
            key_identifier=extension.key_identifier,
            authority_cert_issuer=issuers,
            authority_cert_serial_number=extension.authority_cert_serial_number,
            _x509_obj=extension,
        )

    def _string_dict(self):
        ret = {}

        if self.key_identifier is not None:
            hex_key = _byte_to_hex(self.key_identifier)
            ret["Key Identifier"] = hex_key
        if self.authority_cert_issuer is not None:
            ret["Authority Cert Issuer"] = self.authority_cert_issuer
        if self.authority_cert_serial_number is not None:
            ret[
                "Authority Cert Serial Number"
            ] = self.authority_cert_serial_number

        if ret:
            return {self.name: ret}
        return {}

    def _to_cryptography(self) -> x509.AuthorityKeyIdentifier:
        authority_cert_issuer = []
        for general_name in self.authority_cert_issuer:
            authority = getattr(GENERAL_NAME_MODULE, general_name.name)(
                general_name.value
            )
            authority_cert_issuer.append(authority)

        return x509.AuthorityKeyIdentifier(
            key_identifier=self.key_identifier,
            authority_cert_issuer=authority_cert_issuer,
            authority_cert_serial_number=self.authority_cert_serial_number,
        )


class SubjectKeyIdentifier(Extension):
    """
    Represents a subject key identifier extension.

    Attributes:
        subject_key_identifier: The subject key identifier.
    """

    subject_key_identifier: bytes

    @classmethod
    def from_cryptography(
        cls: Type["SubjectKeyIdentifier"], extension: x509.SubjectKeyIdentifier
    ) -> "SubjectKeyIdentifier":
        """
        Create a SubjectKeyIdentifier instance from a cryptography
        SubjectKeyIdentifier object.

        Args:
            extension: The cryptography SubjectKeyIdentifier object.

        Returns:
            The SubjectKeyIdentifier instance.
        """
        return cls(
            subject_key_identifier=extension.key_identifier,
            _x509_obj=extension,
        )

    def _string_dict(self):
        hex_key = _byte_to_hex(self.subject_key_identifier)
        return {self.name: {"Subject Key Identifier": hex_key}}

    def _to_cryptography(self) -> x509.SubjectKeyIdentifier:
        return x509.SubjectKeyIdentifier(self.subject_key_identifier)


class KeyUsage(Extension):
    """
    Represents a key usage extension.

    Attributes:
        digital_signature: Indicates if digital signature is allowed.
        content_commitment: Indicates if content commitment is allowed.
        key_encipherment: Indicates if key encipherment is allowed.
        data_encipherment: Indicates if data encipherment is allowed.
        key_agreement: Indicates if key agreement is allowed.
        key_cert_sign: Indicates if key certificate signing is allowed.
        crl_sign: Indicates if CRL signing is allowed.
        encipher_only: Indicates if encipher only is allowed.
        decipher_only: Indicates if decipher only is allowed.
    """

    digital_signature: bool = False
    content_commitment: bool = False
    key_encipherment: bool = False
    data_encipherment: bool = False
    key_agreement: bool = False
    key_cert_sign: bool = False
    crl_sign: bool = False
    encipher_only: bool = False
    decipher_only: bool = False

    @classmethod
    def from_cryptography(
        cls: Type["KeyUsage"], extension: x509.KeyUsage
    ) -> "KeyUsage":
        """
        Create a KeyUsage instance from a cryptography KeyUsage object.

        Args:
            extension: The cryptography KeyUsage object.

        Returns:
            The KeyUsage instance.
        """
        try:
            encipher_only = extension.encipher_only
        except ValueError:
            encipher_only = False
        try:
            decipher_only = extension.decipher_only
        except ValueError:
            decipher_only = False

        return cls(
            digital_signature=extension.digital_signature,
            content_commitment=extension.content_commitment,
            key_encipherment=extension.key_encipherment,
            data_encipherment=extension.data_encipherment,
            key_agreement=extension.key_agreement,
            key_cert_sign=extension.key_cert_sign,
            crl_sign=extension.crl_sign,
            encipher_only=encipher_only,
            decipher_only=decipher_only,
            _x509_obj=extension,
        )

    def _string_dict(self):
        true_fields = []
        for field in self.model_fields:
            if field == "critical":
                continue

            field_title = " ".join(ele.title() for ele in field.split("_"))
            if getattr(self, field):
                true_fields.append(field_title)

        return {self.name: ", ".join(true_fields)}

    def _to_cryptography(self) -> x509.KeyUsage:
        return x509.KeyUsage(
            digital_signature=self.digital_signature,
            content_commitment=self.content_commitment,
            key_encipherment=self.key_encipherment,
            data_encipherment=self.data_encipherment,
            key_agreement=self.key_agreement,
            key_cert_sign=self.key_cert_sign,
            crl_sign=self.crl_sign,
            encipher_only=self.encipher_only,
            decipher_only=self.decipher_only,
        )


class NoticeReference(Extension):
    """
    Represents a notice reference extension.

    Attributes:
        organization: The organization associated with the notice.
        notice_numbers: List of notice numbers.
    """

    organization: str
    notice_numbers: List[int]

    @classmethod
    def from_cryptography(cls, notice_reference: x509.NoticeReference):
        return cls(
            organization=notice_reference.organization,
            notice_numbers=notice_reference.notice_numbers,
            _x509_obj=notice_reference,
        )

    def _string_dict(self):
        return {
            self.name: {
                "Organization": self.organization,
                "Notice Numbers": self.notice_numbers,
            }
        }

    def _to_cryptography(self) -> x509.NoticeReference:
        return x509.NoticeReference(
            organization=self.organization,
            notice_numbers=self.notice_numbers,
        )


class UserNotice(Extension):
    """
    Represents a user notice extension.

    Attributes:
        notice_reference: The notice reference associated with the user notice.
        explicit_text: The explicit text of the user notice.
    """

    notice_reference: Optional[NoticeReference]
    explicit_text: Optional[str]

    @classmethod
    def from_cryptography(
        cls: Type["UserNotice"], policy_info: x509.UserNotice
    ) -> "UserNotice":
        """
        Create a UserNotice instance from a cryptography UserNotice object.

        Args:
            policy_info: The cryptography UserNotice object.

        Returns:
            The UserNotice instance.
        """
        return cls(
            notice_reference=NoticeReference.from_cryptography(
                policy_info.notice_reference
            ),
            explicit_text=policy_info.explicit_text,
            _x509_obj=policy_info,
        )

    def _string_dict(self):
        ret = {self.name: {}}

        if self.notice_reference is not None:
            ret[self.name].update(self.notice_reference._string_dict())
        if self.explicit_text is not None:
            ret[self.name]["Explicit Text"] = self.explicit_text
        return ret

    def _to_cryptography(self) -> x509.UserNotice:
        return x509.UserNotice(
            notice_reference=self.notice_reference._to_cryptography(),
            explicit_text=self.explicit_text,
        )


class PolicyInformation(Extension):
    """
    Represents a policy information extension.

    Attributes:
        policy_identifier: The policy identifier associated with the policy
            information.
        policy_qualifiers: List of policy qualifiers.
    """

    policy_identifier: str
    policy_qualifiers: Optional[List[Union[str, UserNotice]]]

    @classmethod
    def from_cryptography(
        cls: Type["PolicyInformation"], policy_info: x509.PolicyInformation
    ) -> "PolicyInformation":
        """
        Create a PolicyInformation instance from a cryptography
        PolicyInformation object.

        Args:
            policy_info: The cryptography PolicyInformation object.

        Returns:
            The PolicyInformation instance.
        """
        policy_qualifiers = None
        if policy_info.policy_qualifiers is not None:
            policy_qualifiers = []
            for qualifier in policy_info.policy_qualifiers:
                if isinstance(qualifier, str):
                    policy_qualifiers.append(qualifier)
                else:
                    policy_qualifiers.append(
                        UserNotice.from_cryptography(qualifier)
                    )

        return cls(
            policy_identifier=policy_info.policy_identifier.dotted_string,
            policy_qualifiers=policy_qualifiers,
            _x509_obj=policy_info,
        )

    def _string_dict(self):
        name = f"Policy {self.policy_identifier}"
        ret = {name: []}

        if self.policy_qualifiers is not None:
            for qualifier in self.policy_qualifiers:
                if isinstance(qualifier, str):
                    ret[name].append(f"CPS: {qualifier}")
                else:
                    ret[name].append(qualifier._string_dict())
        return ret

    def _to_cryptography(self) -> x509.PolicyInformation:
        qualifiers = []
        for qualifier in self.policy_qualifiers:
            if isinstance(qualifier, str):
                qualifiers.append(qualifier)
            else:
                qualifiers.append(qualifier._to_cryptography())

        return x509.PolicyInformation(
            policy_identifier=x509.ObjectIdentifier(self.policy_identifier),
            policy_qualifiers=qualifiers,
        )


class CertificatePolicies(Extension):
    """
    Represents a certificate policies extension.

    Attributes:
        policy_information: List of policy information.
    """

    policy_information: List[PolicyInformation]

    @classmethod
    def from_cryptography(
        cls: Type["CertificatePolicies"], extension: x509.CertificatePolicies
    ) -> "CertificatePolicies":
        """
        Create a CertificatePolicies instance from a cryptography
        CertificatePolicies object.

        Args:
            extension: The cryptography CertificatePolicies object.

        Returns:
            The CertificatePolicies instance.
        """
        res = []
        for policy_information in extension:
            info = PolicyInformation.from_cryptography(policy_information)
            res.append(info)

        return cls(
            policy_information=res,
            _x509_obj=extension,
        )

    def _string_dict(self):
        ret = {self.name: []}
        for policy_information in self.policy_information:
            ret[self.name].append(policy_information._string_dict())

        return ret

    def _to_cryptography(self) -> x509.CertificatePolicies:
        policies = []
        for policy in self.policy_information:
            policies.append(policy._to_cryptography())

        return x509.CertificatePolicies(policies)


class AlternativeName(Extension):
    """
    Represents an alternative name extension.

    Attributes:
        general_names: List of general names.
    """

    general_names: List[GeneralName]

    @classmethod
    def from_cryptography(
        cls: Type["AlternativeName"],
        extension: Union[
            x509.SubjectAlternativeName, x509.IssuerAlternativeName
        ],
    ) -> "AlternativeName":
        """
        Create an AlternativeName instance from a cryptography
        SubjectAlternativeName or IssuerAlternativeName object.

        Args:
            extension: The cryptography SubjectAlternativeName or
                IssuerAlternativeName object.

        Returns:
            The AlternativeName instance.
        """
        names = []
        for general_name in extension:
            names.append(GeneralName.from_cryptography(general_name))

        return cls(
            general_names=names,
            _x509_obj=extension,
        )

    def _string_dict(self):
        general_names = []
        for general_name in self.general_names:
            general_names.append(general_name._string_dict())
        return {self.name: general_names}

    def _to_cryptography(
        self,
    ) -> Union[x509.SubjectAlternativeName, x509.IssuerAlternativeName]:
        general_names = []
        for general_name in self.general_names:
            general_names.append(general_name._to_cryptography())

        subclass_name = type(self).__name__

        return getattr(EXTENSIONS_MODULE, subclass_name)(general_names)


class SubjectAlternativeName(AlternativeName):
    """
    Represents a subject alternative name extension.
    """

    pass


class IssuerAlternativeName(AlternativeName):
    """
    Represents an issuer alternative name extension.
    """

    pass


class SubjectDirectoryAttributes(Extension):
    """
    Represents a subject directory attributes extension.

    Attributes:
        attributes: List of attributes.
    """

    attributes: List[str]

    @classmethod
    def from_cryptography(
        cls: Type["SubjectDirectoryAttributes"],
        extension: x509.UnrecognizedExtension,
    ) -> "SubjectDirectoryAttributes":
        """
        Create a SubjectDirectoryAttributes instance from a cryptography
        UnrecognizedExtension object.

        Args:
            extension: The cryptography UnrecognizedExtension object.

        Returns:
            The SubjectDirectoryAttributes instance.
        """
        attributes = []
        vals = extension.value
        if not isinstance(vals, List):
            vals = [vals]

        for val in vals:
            if isinstance(val, bytes):
                val = _byte_to_hex(val)

            attributes.append(val)

        return cls(attributes=attributes, _x509_obj=extension)

    def _to_cryptography(self) -> x509.UnrecognizedExtension:
        values = []
        for attribute in self.attributes:
            try:
                attribute = _hex_to_byte(attribute)
            except Exception:
                pass

            values.append(attribute)

        if len(values) == 1:
            values = values[0].encode()
        else:
            values = bytes(values)
        return x509.UnrecognizedExtension(
            oid=ObjectIdentifier("2.5.29.9"), value=values
        )

    def _string_dict(self):
        return {self.name: self.attributes}


class BasicConstraints(Extension):
    """
    Represents a basic constraints extension.

    Attributes:
        ca: Indicates if the certificate is a CA.
        path_len_constrain: The path length constraint.
    """

    ca: bool
    path_len_constraint: Optional[int] = None

    @classmethod
    def from_cryptography(
        cls: Type["BasicConstraints"], extension: x509.BasicConstraints
    ) -> "BasicConstraints":
        """
        Create a BasicConstraints instance from a cryptography
        BasicConstraints object.

        Args:
            extension: The cryptography BasicConstraints object.

        Returns:
            The BasicConstraints instance.
        """
        return cls(
            ca=extension.ca,
            path_len_constraint=extension.path_length,
            _x509_obj=extension,
        )

    def _string_dict(self):
        ret = {
            self.name: {
                "CA": self.ca,
            }
        }

        if self.path_len_constraint is not None:
            ret["Path Lenght"] = self.path_len_constraint

        return ret

    def _to_cryptography(self) -> x509.BasicConstraints:
        return x509.BasicConstraints(
            ca=self.ca,
            path_length=self.path_len_constraint,
        )


class NameConstraints(Extension):
    """
    Represents a name constraints extension.

    Attributes:
        permitted_subtrees: List of permitted subtrees.
        excluded_subtrees: List of excluded subtrees.
    """

    permitted_subtrees: Optional[List[GeneralName]]
    excluded_subtrees: Optional[List[GeneralName]]

    @classmethod
    def from_cryptography(cls, extension: x509.NameConstraints):
        permitted_subtrees = []
        for permitted in extension.permitted_subtrees:
            permitted_subtrees.append(GeneralName.from_cryptography(permitted))
        if not permitted_subtrees:
            permitted_subtrees = None

        excluded_subtrees = []
        for excluded in extension.excluded_subtrees:
            excluded_subtrees.append(GeneralName.from_cryptography(excluded))
        if not excluded_subtrees:
            excluded_subtrees = None

        return cls(
            permitted_subtrees=permitted_subtrees,
            excluded_subtrees=excluded_subtrees,
            _x509_obj=extension,
        )

    def _string_dict(self):
        permitted_subtrees = []
        for permitted_subtree in self.permitted_subtrees:
            permitted_subtrees.append(permitted_subtree._string_dict())

        excluded_subtrees = []
        for excluded_subtree in self.excluded_subtrees:
            excluded_subtrees.append(excluded_subtree._string_dict())

        return {
            self.name: {
                "Permitted Subtrees": permitted_subtrees,
                "Excluded Subtrees": excluded_subtrees,
            }
        }

    def _to_cryptography(self) -> x509.NameConstraints:
        permitted_subtrees = None
        if self.permitted_subtrees is not None:
            permitted_subtrees = []
            for permitted_subtree in self.permitted_subtrees:
                permitted_subtrees.append(permitted_subtree._to_cryptography())

        excluded_subtrees = None
        if self.excluded_subtrees is not None:
            excluded_subtrees = []
            for excluded_subtree in self.excluded_subtrees:
                excluded_subtrees.append(excluded_subtree._to_cryptography())

        return x509.NameConstraints(
            permitted_subtrees=permitted_subtrees,
            excluded_subtrees=excluded_subtrees,
        )


class PolicyConstraints(Extension):
    """
    Represents a policy constraints extension.

    Attributes:
        require_explicit_policy: The require explicit policy value.
        inhibit_policy_mapping: The inhibit policy mapping value.
    """

    require_explicit_policy: Optional[int]
    inhibit_policy_mapping: Optional[int]

    @classmethod
    def from_cryptography(cls, extension: x509.PolicyConstraints):
        return cls(
            require_explicit_policy=extension.require_explicit_policy,
            inhibit_policy_mapping=extension.inhibit_policy_mapping,
            _x509_obj=extension,
        )

    def _string_dict(self):
        ret = {self.name: {}}

        if self.require_explicit_policy is not None:
            ret[self.name][
                "Require Explicit Policy"
            ] = self.require_explicit_policy
        if self.inhibit_policy_mapping is not None:
            ret[self.name][
                "Inhibit Policy Mapping"
            ] = self.inhibit_policy_mapping

        return ret

    def _to_cryptography(self) -> x509.PolicyConstraints:
        return x509.PolicyConstraints(
            require_explicit_policy=self.require_explicit_policy,
            inhibit_policy_mapping=self.inhibit_policy_mapping,
        )


EKU_OID_MAPPING = {
    "1.3.6.1.5.5.7.3.1": "Server Authentication",
    "1.3.6.1.5.5.7.3.2": "Client Authentication",
    "1.3.6.1.5.5.7.3.3": "Code Signing",
    "1.3.6.1.5.5.7.3.4": "Email Protection",
    "1.3.6.1.5.5.7.3.8": "Time Stamping",
    "1.3.6.1.5.5.7.3.9": "OCSP Signing",
    "1.3.6.1.5.5.7.3.17": "IPsec IKE",
    "2.5.29.37.0": "Any Extended Key Usage",
    "1.3.6.1.4.1.311.20.2.2": "Smart Card Logon",
    "1.3.6.1.5.2.3.5": "Kerberos PKINIT KDC",
    "1.3.6.1.4.1.11129.2.4.4": "Certificate Transparency",
}


class ExtendedKeyUsage(Extension):
    """
    Represents the Extended Key Usage extension in X.509 certificates.

    Attributes:
        ext_key_usage_syntax: List of extended key usage OIDs.
    """

    ext_key_usage_syntax: List[str]

    @classmethod
    def from_cryptography(
        cls: Type["ExtendedKeyUsage"], extension: x509.ExtendedKeyUsage
    ) -> "ExtendedKeyUsage":
        """
        Constructs an ExtendedKeyUsage object from a cryptography
        ExtendedKeyUsage object.

        Args:
            extension: The cryptography ExtendedKeyUsage object.

        Returns:
            The constructed ExtendedKeyUsage object.
        """
        ext_key_usage_syntax = []
        for key_usage in extension:
            ext_key_usage_syntax.append(key_usage.dotted_string)
        return cls(
            ext_key_usage_syntax=ext_key_usage_syntax,
            _x509_obj=extension,
        )

    def _string_dict(self):
        names = []
        for oid in self.ext_key_usage_syntax:
            names.append(
                EKU_OID_MAPPING.get(
                    oid,
                    f"Unknown OID ({oid})",
                )
            )

        return {self.name: names}

    def _to_cryptography(self) -> x509.ExtendedKeyUsage:
        oids = []
        for oid in self.ext_key_usage_syntax:
            oids.append(x509.ObjectIdentifier(oid))
        return x509.ExtendedKeyUsage(oids)


class AttributeTypeAndValue(CryptoParser):
    """
    Represents an Attribute Type and Value in X.509 certificates.

    Attributes:
        oid: The Object Identifier (OID) of the attribute.
        value: The value of the attribute.
    """

    oid: str
    value: str

    @classmethod
    def from_cryptography(
        cls: Type["AttributeTypeAndValue"], x509_obj: x509.NameAttribute
    ) -> "AttributeTypeAndValue":
        """
        Constructs an AttributeTypeAndValue object from a cryptography
        NameAttribute object.

        Args:
            x509_obj: The cryptography NameAttribute object.

        Returns:
            The constructed AttributeTypeAndValue object.
        """
        return cls(
            oid=x509_obj.oid.dotted_string,
            value=x509_obj.value,
            _x509_obj=x509_obj,
        )

    def _to_cryptography(self) -> x509.NameAttribute:
        return x509.NameAttribute(
            oid=x509.ObjectIdentifier(self.oid),
            value=self.value,
        )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "OID": self.oid,
            "Value": self.value,
        }


class RelativeDistinguishedName(CryptoParser):
    """
    Represents a Relative Distinguished Name in X.509 certificates.

    Attributes:
        attributes: List of attributes in the RDN.
    """

    attributes: List[AttributeTypeAndValue]

    def __iter__(self) -> Iterable:
        return iter(self.attributes)

    @classmethod
    def from_cryptography(
        cls: Type["RelativeDistinguishedName"],
        x509_obj: x509.RelativeDistinguishedName,
    ) -> "RelativeDistinguishedName":
        """
        Constructs a RelativeDistinguishedName object from a cryptography
        RelativeDistinguishedName object.

        Args:
            x509_obj: The cryptography RelativeDistinguishedName object.

        Returns:
            The constructed RelativeDistinguishedName object.
        """
        attributes = []
        for name_attribute in x509_obj:
            attributes.append(
                AttributeTypeAndValue.from_cryptography(name_attribute)
            )

        cls(attributes=attributes, _x509_obj=x509_obj)

    def _string_dict(self) -> typing.Dict:
        attributes = []
        for att in self.attributes:
            attributes.append(att._string_dict())

        return {"RelativeDistinguishedName": attributes}

    def _to_cryptography(self) -> x509.RelativeDistinguishedName:
        name_attributes = []
        for att in self.attributes:
            name_attributes.append(att._to_cryptography())
        return x509.RelativeDistinguishedName(name_attributes)


class Reason(Enum):
    """
    Reasons for the DistributionPoint extension
    """

    unspecified = "unspecified"
    key_compromise = "key_compromise"
    ca_compromise = "ca_compromise"
    affiliation_changed = "affiliation_changed"
    superseded = "superseded"
    cessation_of_operation = "cessation_of_operation"
    certificate_hold = "certificate_hold"
    privilege_withdrawn = "privilege_withdrawn"
    aa_compromise = "aa_compromise"
    remove_from_crl = "remove_from_crl"


class DistributionPoint(CryptoParser):
    """
    Represents a Distribution Point in X.509 certificates.

    Attributes:
        full_name: List of full names associated with the distribution point.
        name_relative_to_crl_issuer: Relative name to the CRL issuer.
        reasons: List of reasons for the distribution point.
        crl_issuer: List of CRL issuers associated with the distribution
            point.
    """

    full_name: Optional[List[GeneralName]] = None
    name_relative_to_crl_issuer: Optional[RelativeDistinguishedName] = None
    reasons: Optional[List[Reason]] = None
    crl_issuer: Optional[List[GeneralName]] = None

    @classmethod
    def from_cryptography(
        cls: Type["DistributionPoint"], extension: x509.DistributionPoint
    ) -> "DistributionPoint":
        """
        Constructs a DistributionPoint object from a cryptography
        DistributionPoint object.

        Args:
            extension: The cryptography DistributionPoint object.

        Returns:
            DistributionPoint: The constructed DistributionPoint object.
        """
        full_names = None
        if extension.full_name is not None:
            full_names = []
            for full_name in extension.full_name:
                full_names.append(GeneralName.from_cryptography(full_name))

        relative_name = None
        if extension.relative_name is not None:
            relative_name = RelativeDistinguishedName.from_cryptography(
                extension.relative_name
            )

        reasons = None
        if extension.reasons is not None:
            reasons = []
            for reason in extension.reasons:
                reasons.append(getattr(Reason, reason.name))

        crl_issuers = None
        if extension.crl_issuer is not None:
            crl_issuers = []
            for crl_issuer in extension.crl_issuer:
                crl_issuers.append(GeneralName.from_cryptography(crl_issuer))

        return cls(
            full_name=full_names,
            name_relative_to_crl_issuer=relative_name,
            reasons=reasons,
            crl_issuer=crl_issuers,
            _x509_obj=extension,
        )

    def _string_dict(self):
        ret = {}
        if self.full_name is not None:
            ret["Full Name"] = []
            for full_name in self.full_name:
                ret["Full Name"].append(full_name._string_dict())
        if self.name_relative_to_crl_issuer is not None:
            ret[
                "Name Relative To CRL Issuer"
            ] = self.name_relative_to_crl_issuer
        if self.reasons is not None:
            ret["Reasons"] = []
            for reason in self.reasons:
                ret["Reasons"].append(
                    getattr(x509.ReasonFlags, reason.name).value
                )
        if self.crl_issuer is not None:
            ret["CRL Issuer"] = self.crl_issuer
        return ret

    def _to_cryptography(self) -> x509.DistributionPoint:
        full_names = None
        if self.full_name is not None:
            full_names = []
            for full_name in self.full_name:
                full_names.append(full_name._to_cryptography())

        relative_names = None
        if self.name_relative_to_crl_issuer is not None:
            relative_names = (
                self.name_relative_to_crl_issuer._to_cryptography()
            )

        reasons = None
        if self.reasons is not None:
            reasons = []
            for reason in self.reasons:
                reasons.append(getattr(x509.ReasonFlags, reason.name))
            reasons = frozenset(reasons)

        crl_issuers = None
        if self.crl_issuer is not None:
            crl_issuers = []
            for crl_issuer in self.crl_issuer:
                crl_issuers.append(crl_issuer._to_cryptography())

        return x509.DistributionPoint(
            full_name=full_names,
            relative_name=relative_names,
            reasons=reasons,
            crl_issuer=crl_issuers,
        )


class CrlDistributionPoints(Extension):
    """
    Represents the CRL Distribution Points extension in X.509 certificates.

    Attributes:
        crl_distribution_points: List of CRL distribution points.
    """

    crl_distribution_points: List[DistributionPoint]

    def __iter__(self) -> Iterable[DistributionPoint]:
        return iter(self.crl_distribution_points)

    @classmethod
    def from_cryptography(
        cls: Type["CrlDistributionPoints"],
        extension: x509.CRLDistributionPoints,
    ) -> "CrlDistributionPoints":
        """
        Constructs a CrlDistributionPoints object from a cryptography
        CRLDistributionPoints object.

        Args:
            extension: The cryptography CRLDistributionPoints object.

        Returns:
            The constructed CrlDistributionPoints object.
        """
        crl_distribution_points = []
        for crl_distribution_point in extension:
            parsed = DistributionPoint.from_cryptography(
                crl_distribution_point
            )
            crl_distribution_points.append(parsed)

        return cls(
            crl_distribution_points=crl_distribution_points,
            _x509_obj=extension,
        )

    def _string_dict(self):
        ret = {self.name: []}
        for dist_point in self.crl_distribution_points:
            ret[self.name].append(dist_point._string_dict())

        return ret

    def _to_cryptography(self) -> x509.CRLDistributionPoints:
        dist_points = []
        for dist_point in self.crl_distribution_points:
            dist_points.append(dist_point._to_cryptography())

        return x509.CRLDistributionPoints(dist_points)


class InhibitAnyPolicy(Extension):
    """
    Represents the Inhibit Any Policy extension in X.509 certificates.

    Attributes:
        skip_certs: Number of certificates to skip.
    """

    skip_certs: int

    @classmethod
    def from_cryptography(
        cls: Type["InhibitAnyPolicy"], extension: x509.InhibitAnyPolicy
    ) -> "InhibitAnyPolicy":
        """
        Constructs an InhibitAnyPolicy object from a cryptography
        InhibitAnyPolicy object.

        Args:
            extension: The cryptography InhibitAnyPolicy object.

        Returns:
            The constructed InhibitAnyPolicy object.
        """
        return cls(
            skip_certs=extension.skip_certs,
            _x509_obj=extension,
        )

    def _string_dict(self):
        return {self.name: {"Skip Certs": self.skip_certs}}

    def _to_cryptography(self) -> x509.InhibitAnyPolicy:
        return x509.InhibitAnyPolicy(self.skip_certs)


class FreshestCrl(CrlDistributionPoints):
    """
    Represents the Freshest CRL extension in X.509 certificates.
    """

    def _to_cryptography(self) -> x509.FreshestCRL:
        dist_points = []
        for dist_point in self.crl_distribution_points:
            dist_points.append(dist_point._to_cryptography())

        return x509.FreshestCRL(dist_points)


class AccessDescriptionId(Enum):
    """
    Enumeration of access description identifiers for X.509 certificates.
    """

    CA_ISSUERS = "1.3.6.1.5.5.7.48.2"
    OCSP = "1.3.6.1.5.5.7.48.1"


class AccessDescription(CryptoParser):
    """
    Represents an Access Description in X.509 certificates.

    Attributes:
        access_method: The access method identifier.
        access_location: The access location.
    """

    access_method: AccessDescriptionId
    access_location: GeneralName

    @classmethod
    def from_cryptography(
        cls: Type["AccessDescription"], extension: x509.AccessDescription
    ) -> "AccessDescription":
        """
        Constructs an AccessDescription object from a cryptography
        AccessDescription object.

        Args:
            extension: The cryptography AccessDescription object.

        Returns:
            The constructed AccessDescription object.
        """
        access_method = AccessDescriptionId(
            extension.access_method.dotted_string
        )
        return cls(
            access_method=access_method,
            access_location=GeneralName.from_cryptography(
                extension.access_location
            ),
            _x509_obj=extension,
        )

    def _string_dict(self):
        return {
            "Access Method": self.access_method.name,
            "Access Location": self.access_location._string_dict(),
        }

    def _to_cryptography(self) -> x509.AccessDescription:
        return x509.AccessDescription(
            access_method=getattr(
                AuthorityInformationAccessOID, self.access_method.name
            ),
            access_location=self.access_location._to_cryptography(),
        )


class AuthorityInformationAccess(Extension):
    """
    Represents the Authority Information Access extension in X.509
    certificates.

    Attributes:
        access_description: List of access descriptions.
    """

    access_description: List[AccessDescription]

    def __iter__(self) -> Iterable[AccessDescription]:
        return iter(self.access_description)

    @classmethod
    def from_cryptography(
        cls: Type["AuthorityInformationAccess"],
        extension: x509.AuthorityInformationAccess,
    ) -> "AuthorityInformationAccess":
        """
        Constructs an AuthorityInformationAccess object from a
        cryptography AuthorityInformationAccess object.

        Args:
            extension: The cryptography AuthorityInformationAccess object.

        Returns:
            The constructed AuthorityInformationAccess object.
        """
        access_descriptions = []
        for access_description in extension:
            access_descriptions.append(
                AccessDescription.from_cryptography(access_description)
            )
        return cls(
            access_description=access_descriptions,
            _x509_obj=extension,
        )

    def _string_dict(self):
        access_descriptions = []
        for access_description in self.access_description:
            access_descriptions.append(access_description._string_dict())

        return {self.name: {"Access Description": access_descriptions}}

    def _to_cryptography(self) -> x509.AuthorityInformationAccess:
        access_descriptions = []
        for access_description in self.access_description:
            access_descriptions.append(access_description._to_cryptography())
        return x509.AuthorityInformationAccess(access_descriptions)


class SubjectInformationAccess(AuthorityInformationAccess):
    """
    Represents the Subject Information Access extension in X.509
    certificates.
    """

    def _to_cryptography(self) -> x509.SubjectInformationAccess:
        access_descriptions = []
        for access_description in self.access_description:
            access_descriptions.append(access_description._to_cryptography())
        return x509.SubjectInformationAccess(access_descriptions)


class Extensions(CryptoParser):
    """
    Extensions in X.509 certificates.

    Attributes:
        authority_key_identifier: Authority Key Identifier extension.
        subject_key_identifier: Subject Key Identifier extension.
        key_usage: Key Usage extension.
        certificate_policies: Certificate Policies extension.
        subject_alternative_name: Subject Alternative Name extension.
        issuer_alternative_name: Issuer Alternative Name extension.
        subject_directory_attributes: Subject Directory Attributes extension.
        basic_constraints: Basic Constraints extension.
        name_constraints: Name Constraints extension.
        policy_constraints: Policy Constraints extension.
        extended_key_usage: Extended Key Usage extension.
        crl_distribution_points: CRL Distribution Points extension.
        inhibit_any_policy: Inhibit Any Policy extension.
        freshest_crl: Freshest CRL extension.
        authority_information_access: Authority Information Access extension.
        subject_information_access: Subject Information Access extension.
    """

    model_config = ConfigDict(populate_by_name=True)

    authority_key_identifier: Optional[AuthorityKeyIdentifier] = Field(
        alias=ExtensionOID.AUTHORITY_KEY_IDENTIFIER.dotted_string, default=None
    )
    subject_key_identifier: Optional[SubjectKeyIdentifier] = Field(
        alias=ExtensionOID.SUBJECT_KEY_IDENTIFIER.dotted_string, default=None
    )
    key_usage: Optional[KeyUsage] = Field(
        alias=ExtensionOID.KEY_USAGE.dotted_string, default=None
    )
    certificate_policies: Optional[CertificatePolicies] = Field(
        alias=ExtensionOID.CERTIFICATE_POLICIES.dotted_string, default=None
    )
    # policy_mappings
    subject_alternative_name: Optional[SubjectAlternativeName] = Field(
        alias=ExtensionOID.SUBJECT_ALTERNATIVE_NAME.dotted_string, default=None
    )
    issuer_alternative_name: Optional[IssuerAlternativeName] = Field(
        alias=ExtensionOID.ISSUER_ALTERNATIVE_NAME.dotted_string, default=None
    )
    subject_directory_attributes: Optional[SubjectDirectoryAttributes] = Field(
        alias=ExtensionOID.SUBJECT_DIRECTORY_ATTRIBUTES.dotted_string,
        default=None,
    )
    basic_constraints: Optional[BasicConstraints] = Field(
        alias=ExtensionOID.BASIC_CONSTRAINTS.dotted_string,
        default=None,
    )
    name_constraints: Optional[NameConstraints] = Field(
        alias=ExtensionOID.NAME_CONSTRAINTS.dotted_string,
        default=None,
    )
    policy_constraints: Optional[PolicyConstraints] = Field(
        alias=ExtensionOID.POLICY_CONSTRAINTS.dotted_string,
        default=None,
    )
    extended_key_usage: Optional[ExtendedKeyUsage] = Field(
        alias=ExtensionOID.EXTENDED_KEY_USAGE.dotted_string,
        default=None,
    )
    crl_distribution_points: Optional[CrlDistributionPoints] = Field(
        alias=ExtensionOID.CRL_DISTRIBUTION_POINTS.dotted_string,
        default=None,
    )
    inhibit_any_policy: Optional[InhibitAnyPolicy] = Field(
        alias=ExtensionOID.INHIBIT_ANY_POLICY.dotted_string,
        default=None,
    )
    freshest_crl: Optional[FreshestCrl] = Field(
        alias=ExtensionOID.FRESHEST_CRL.dotted_string,
        default=None,
    )
    authority_information_access: Optional[AuthorityInformationAccess] = Field(
        alias=ExtensionOID.AUTHORITY_INFORMATION_ACCESS.dotted_string,
        default=None,
    )
    subject_information_access: Optional[SubjectInformationAccess] = Field(
        alias=ExtensionOID.SUBJECT_INFORMATION_ACCESS.dotted_string,
        default=None,
    )

    def __iter__(self) -> Iterable[Extension]:
        for field_name, field in self.model_fields.items():
            val = getattr(self, field_name)
            if val is None:
                continue

            yield getattr(self, field_name)

    @classmethod
    def from_cryptography(
        cls: Type["Extensions"], cert_extensions: x509.Extensions
    ) -> "Extensions":
        """
        Constructs an Extensions object from cryptography X.509 Extensions.

        Args:
            cert_extensions: The cryptography X.509 Extensions.

        Returns:
            The constructed Extensions object.
        """
        extensions_dict = {"_x509_obj": cert_extensions}

        for name, field_info in cls.model_fields.items():
            class_type = typing.get_args(field_info.annotation)[0]
            oid = ObjectIdentifier(field_info.alias)
            try:
                x509_ext = cert_extensions.get_extension_for_oid(oid)

                ext = class_type.from_cryptography(x509_ext.value)

                extensions_dict[oid.dotted_string] = ext
                extensions_dict[oid.dotted_string].critical = x509_ext.critical
            except ExtensionNotFound:
                logger.trace(f"Extension with OID: {oid._name} not found")

        return cls(**extensions_dict)

    @staticmethod
    def _get_extension_from_oid(
        cert_extensions: x509.Extensions, oid: ObjectIdentifier, classType
    ) -> Optional[ExtensionTypeVar]:
        try:
            ext_val = cert_extensions.get_extension_for_oid(oid).value
            classType.from_cryptography(ext_val)
        except ExtensionNotFound:
            logger.debug(f"Extension with OID: {oid._name} not found")
            return None

    def _string_dict(self):
        extensions = {}
        for field_name in self.model_fields:
            att_val = getattr(self, field_name)

            if att_val is None or str(att_val) == "":
                continue

            extensions.update(att_val._string_dict())

        return extensions

    def _to_cryptography(self) -> x509.Extensions:
        extensions = []
        for field_name in self.model_fields:
            att_val = getattr(self, field_name)

            if att_val is None or str(att_val) == "":
                continue

            extensions.append(att_val._to_cryptography())

        return x509.Extensions(extensions)



from typing import Type, Optional, Dict
import re

import yaml
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from loguru import logger

from pki_tools.exceptions import CsrLoadError, MissingInit
from pki_tools.types.key_pair import KeyPair, CryptoKeyPair
from pki_tools.types.certificate import _is_pem_csr_string
from pki_tools.types.crypto_parser import InitCryptoParser
from pki_tools.types.name import Name
from pki_tools.types.extensions import Extensions
from pki_tools.types.signature_algorithm import (
    SignatureAlgorithm,
)
from pki_tools.types.utils import _byte_to_hex


class CertificateSigningRequest(InitCryptoParser):
    """
    Represents a certificate signing request (CSR).

    Attributes:
        subject: The subject of the CSR.
        public_key: Public key associated with the CSR.
        extensions: Extensions associated with the CSR.
        attributes: Attributes of the CSR.
        signature_algorithm: Signature algorithm used while signing the CSR.
    """

    subject: Name

    public_key: Optional[KeyPair] = None
    extensions: Optional[Extensions] = None
    attributes: Optional[Dict[str, bytes]] = None
    signature_algorithm: Optional[SignatureAlgorithm] = None

    _private_key: Optional[CryptoKeyPair]

    @classmethod
    def from_cryptography(
        cls: Type["CertificateSigningRequest"],
        crypto_csr: x509.CertificateSigningRequest,
    ) -> "CertificateSigningRequest":
        """
        Create a CertificateSigningRequest object from a cryptography
        CertificateSigningRequest.

        Args:
            crypto_csr: Cryptography CertificateSigningRequest.

        Returns:
            Instance of CertificateSigningRequest.
        """
        attributes = {}
        for att in crypto_csr.attributes:
            attributes[att.oid.dotted_string] = att.value

        ret = cls(
            subject=Name.from_cryptography(crypto_csr.subject),
            extensions=Extensions.from_cryptography(crypto_csr.extensions),
            signature_algorithm=SignatureAlgorithm.from_cryptography(
                crypto_csr.signature_hash_algorithm,
                crypto_csr.signature_algorithm_parameters,
            ),
            signature_value=_byte_to_hex(crypto_csr.signature),
            public_key=KeyPair.from_cryptography(crypto_csr.public_key()),
            attributes=attributes,
            _x509_obj=crypto_csr,
        )
        ret._x509_obj = crypto_csr
        return ret

    @classmethod
    def from_pem_string(
        cls: Type["CertificateSigningRequest"], csr_pem: str
    ) -> "CertificateSigningRequest":
        """
        Load a CSR from a PEM string into a CertificateSigningRequest object.

        Args:
            csr_pem: PEM encoded CSR in string format.

        Returns:
            Instance of CertificateSigningRequest.

        Raises:
            CsrLoadError: If the CSR could not be loaded.
        """
        try:
            csr_pem = re.sub(r"\n\s*", "\n", csr_pem)
            if not _is_pem_csr_string(csr_pem):
                raise ValueError

            csr_cert = x509.load_pem_x509_csr(csr_pem.encode())
            return CertificateSigningRequest.from_cryptography(csr_cert)
        except ValueError as e:
            logger.bind(csr=csr_pem).debug("Failed to load CSR from PEM")
            raise CsrLoadError(e)

    @classmethod
    def from_file(
        cls: Type["CertificateSigningRequest"], file_path: str
    ) -> "CertificateSigningRequest":
        """
        Read a file containing a PEM CSR into a CertificateSigningRequest
        object.

        Args:
            file_path: Path and filename of the PEM CSR.

        Returns:
            Instance of CertificateSigningRequest representing the CSR loaded
            from file.
        """
        with open(file_path, "r") as f:
            csr_pem = f.read()

        return CertificateSigningRequest.from_pem_string(csr_pem)

    @property
    def tbs_bytes(self) -> bytes:
        """
        Get the bytes to be signed of the CSR.

        Returns:
            TBS bytes of the CSR.
        """
        return self._crypto_object.tbs_certrequest_bytes

    @property
    def pem_bytes(self) -> bytes:
        """
        Get the PEM bytes of the CSR.

        Returns:
            PEM bytes of the CSR.
        """
        return self._crypto_object.public_bytes(
            encoding=serialization.Encoding.PEM
        )

    @property
    def pem_string(self) -> str:
        """
        Get the PEM string representation of the CSR.

        Returns:
            PEM string of the CSR.
        """
        return self.pem_bytes.decode()

    def to_file(self, file_path: str) -> None:
        """
        Save the CSR to a file.

        Args:
            file_path: Path to save the file.
        """
        with open(file_path, "w") as f:
            f.write(self.pem_string)

    def sign(
        self, key_pair: CryptoKeyPair, signature_algorithm: SignatureAlgorithm
    ):
        """
        Sign the CSR with the provided key pair and signature algorithm.

        Args:
            key_pair: Key pair with the private key to use while signing the
                CSR
            signature_algorithm: Signature algorithm to use for signing.
        """
        self._private_key = key_pair
        self.signature_algorithm = signature_algorithm
        self._x509_obj = self._to_cryptography()

    def _string_dict(self):
        ret = {
            "Certificate Signing Request": {
                "Subject": self.subject._string_dict(),
                "Extensions": self.extensions._string_dict(),
                "Signature Algorithm": self.signature_algorithm._string_dict(),
                "Public Key": self.public_key._string_dict(),
            }
        }
        if self.attributes is not None:
            attributes = []

            for k in self.attributes:
                val = self.attributes[k]
                if k == "1.2.840.113549.1.9.14":
                    val = _byte_to_hex(val)
                elif k == "1.2.840.113549.1.9.7":
                    val = val.decode()
                attributes.append(f"{k}: {str(val)}")

            ret["Certificate Signing Request"]["Attributes"] = attributes
        return ret

    def _to_cryptography(self) -> x509.CertificateSigningRequest:
        if not hasattr(self, "_private_key"):
            raise MissingInit(
                f"Please use CertificateSigningRequest."
                f"{self._init_func} function"
            )

        crypto_key = self._private_key._to_cryptography()
        if not hasattr(crypto_key, "public_key"):
            raise MissingInit("Use private key not public")

        builder = x509.CertificateSigningRequestBuilder()
        builder = builder.subject_name(self.subject._to_cryptography())

        if self.extensions is not None:
            for extension in self.extensions:
                builder = builder.add_extension(
                    extension._to_cryptography(), extension.critical
                )

        if self.attributes is not None:
            for attribute_oid, value in self.attributes.items():
                oid = x509.ObjectIdentifier(attribute_oid)
                builder = builder.add_attribute(oid, value)

        return builder.sign(
            crypto_key,
            self.signature_algorithm.algorithm._to_cryptography(),
        )

    def __str__(self) -> str:
        return yaml.safe_dump(
            self._string_dict(),
            indent=2,
            default_flow_style=False,
            explicit_start=True,
            default_style="",
        )



import binascii
import socket
import ssl
from functools import lru_cache

import httpx
from cryptography.hazmat.primitives._serialization import (
    Encoding,
    PublicFormat,
)
from loguru import logger
from pydantic import BaseModel, constr

from pki_tools.exceptions import FetchFailure

CACHE_TIME_SECONDS = 60 * 60 * 24 * 30  # 1 month
HTTPX_CLIENT = httpx.Client(
    transport=httpx.HTTPTransport(retries=2), timeout=15
)


class CertsUri(BaseModel):
    """
    Describes a URI where one or more public certificate(s)
    can be downloaded

    Attributes:
        uri: The URI for the public certificate(s)
            cache_time_seconds: Specifies how long the public cert should be
            cached, default is 1 month.
    """

    uri: constr(pattern=r"https*://.*")
    cache_time_seconds: int = CACHE_TIME_SECONDS

    @property
    def hostname(self):
        ret = self.uri.replace("https://", "")
        ret = ret.replace("http://", "")
        return ret


def _byte_to_hex(bytes_in: bytes) -> str:
    return binascii.hexlify(bytes_in).decode().upper()


def _hex_to_byte(hex_string: str) -> bytes:
    byte_array = bytearray()
    for i in range(0, len(hex_string), 2):
        byte_array.append(int(hex_string[i : i + 2], 16))

    return bytes(byte_array)


def _der_key(public_key) -> bytes:
    return public_key.public_bytes(
        encoding=Encoding.DER,
        format=PublicFormat.PKCS1,
    )


@lru_cache(maxsize=None)
def _download_server_certificate(hostname: str, cache_ttl: int = None):
    context = ssl.create_default_context()
    with socket.create_connection((hostname, 443)) as sock:
        with context.wrap_socket(sock, server_hostname=hostname) as ssock:
            der_cert = ssock.getpeercert(binary_form=True)
            return ssl.DER_cert_to_PEM_cert(der_cert)


@lru_cache(maxsize=None)
def _download_pem(uri: str, ttl: int = None) -> str:
    ret = HTTPX_CLIENT.get(uri)

    if ret.status_code != 200:
        logger.bind(status=ret.status_code).error(
            "Failed to fetch issuer from URI"
        )
        raise FetchFailure(f"Failed to fetch URI. Status: {ret.status_code}")

    return ret.text



import yaml

from .certificates import Certificates
from .certificate import Certificate, Validity
from .chain import Chain
from .crl import CertificateRevocationList
from .csr import CertificateSigningRequest
from .extensions import Extensions
from .name import Name
from .enums import RevokeMode
from .key_pair import (
    CryptoKeyPair,
    DSAKeyPair,
    RSAKeyPair,
    EllipticCurveKeyPair,
    Ed25519KeyPair,
    Ed448KeyPair,
    KeyPair,
)
from .signature_algorithm import (
    SignatureAlgorithm,
    HashAlgorithm,
    HashAlgorithmName,
    SHA1,
    SHA512_224,
    SHA512_256,
    SHA224,
    SHA256,
    SHA384,
    SHA512,
    SHA3_224,
    SHA3_256,
    SHA3_384,
    SHA3_512,
    SHAKE128,
    SHAKE256,
    MD5,
    BLAKE2b,
    BLAKE2s,
    SM3,
)

MAX_YAML_LEN = 80


def str_presenter(dumper, data):
    if len(data) > MAX_YAML_LEN:
        chunks = [
            data[i : i + MAX_YAML_LEN]
            for i in range(0, len(data), MAX_YAML_LEN)
        ]

        data = "\n".join(chunks)
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str", data, style="|"
        )
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.representer.SafeRepresenter.add_representer(str, str_presenter)



import time
from functools import lru_cache

from loguru import logger

from pki_tools.types.extensions import UniformResourceIdentifier
from pki_tools.types.utils import HTTPX_CLIENT
from pki_tools.types.chain import Chain
from pki_tools.types.certificate import Certificate
from pki_tools.exceptions import (
    ExtensionMissing,
    CrlFetchFailure,
)
from pki_tools.types.crl import CertificateRevocationList


def _is_revoked(
    cert: Certificate,
    crl_issuer: Chain,
    crl_cache_seconds: int = 3600,
) -> bool:
    crl_issuer.check_chain()
    logger.debug("CRL issuer chain valid")

    log = logger.bind(
        cert=cert.pem_string,
        serial=cert.hex_serial,
    )

    if cert.extensions.crl_distribution_points is None:
        log.debug("CRL extension missing")
        raise ExtensionMissing()

    http_dist = False
    for dist_point in cert.extensions.crl_distribution_points:
        if dist_point.full_name is None:
            continue

        for full_name in dist_point.full_name:
            if not isinstance(full_name, UniformResourceIdentifier):
                logger.warning(
                    "CRL Distribution Point is not "
                    "UniformResourceIdentifier"
                )
                continue

            uri = full_name.value

            http_dist = True
            cache_ttl = round(time.time() / crl_cache_seconds)
            crl = _get_crl_from_url(uri, cache_ttl=cache_ttl)

            issuer = crl_issuer.get_issuer(crl)

            issuer.verify_signature(crl)
            logger.debug("CRL signature valid")

            if (r := crl.get_revoked(cert.serial_number)) is not None:
                log.bind(date=str(r.date)).debug("Certificate revoked")
                return True

    if not http_dist:
        log.debug("CRL missing URI")
        raise ExtensionMissing()

    log.debug("Certificate valid")
    return False


@lru_cache(maxsize=None)
def _get_crl_from_url(crl_url, cache_ttl=None) -> CertificateRevocationList:
    ret = HTTPX_CLIENT.get(crl_url)

    if ret.status_code != 200:
        logger.bind(
            status=ret.status_code,
            url=crl_url,
        ).error("Failed to fetch CRL from URL")
        raise CrlFetchFailure()

    return CertificateRevocationList.from_bytes(ret.content)



import base64
import random
import re
import time
from typing import Optional, Dict
import datetime

import yaml

from cryptography.hazmat.primitives import serialization

from pki_tools.types.key_pair import KeyPair, CryptoKeyPair
from pki_tools.types.name import Name
from pki_tools.types.extensions import Extensions

from pki_tools.exceptions import (
    CertLoadError,
    MissingInit,
    SignatureVerificationFailed,
)
from pki_tools.types.signature_algorithm import (
    SignatureAlgorithm,
    HashAlgorithm,
    HashAlgorithmName,
    PKCS1v15Padding,
)
from pki_tools.types.utils import (
    _byte_to_hex,
    _der_key,
    CertsUri,
    CACHE_TIME_SECONDS,
    _download_server_certificate,
    _download_pem,
)

from typing import Type

from cryptography import x509

from pydantic import BaseModel


from pki_tools.types.crypto_parser import InitCryptoParser

from loguru import logger
from pydantic import ConfigDict


PEM_CERT_REGEX = re.compile(
    r"\s*-+BEGIN CERTIFICATE-+[\w+/\s=]*-+END CERTIFICATE-+\s*"
)
PEM_CSR_REGEX = re.compile(
    r"\s*-+BEGIN CERTIFICATE REQUEST-+[\w+/\s=]*-+END CERTIFICATE REQUEST-+\s*"
)


class Validity(BaseModel):
    """
    Describes the validity of a certificate

    Attributes:
        not_before: The start date of when the certificate will be valid
        not_after: The date of when the certificate expires
    """

    not_before: datetime.datetime
    not_after: datetime.datetime

    def _string_dict(self) -> Dict[str, str]:
        return {
            "Not Before": str(self.not_before),
            "Not After": str(self.not_after),
        }


class Certificate(InitCryptoParser):
    """
    An object describing a x509 Certificate

    Attributes:
        issuer: Certificate issuer
        subject: Certificate subject
        validity: Contains information about NotBefore and NotAfter

        extensions: Certificate (v3) extensions
        serial_number: Serial number
        version: The version of the certificate
        signature_algorithm: Describes the algorithm used to sign the
            certificate
        subject_public_key_info: The public key information
    """

    issuer: Name
    validity: Validity
    subject: Name

    extensions: Optional[Extensions] = None
    serial_number: Optional[int] = None
    version: Optional[int] = None
    signature_algorithm: Optional[SignatureAlgorithm] = None
    subject_public_key_info: Optional[KeyPair] = None

    signature_value: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    _private_key: Optional[CryptoKeyPair]

    @classmethod
    def from_cryptography(
        cls: Type["Certificate"],
        cert: x509.Certificate,
    ) -> "Certificate":
        """
        Create a Certificate object from a [cryptography.x509.Certificate][]
        object.

        Args:
            cert: The [cryptography.x509.Certificate][] object.

        Returns:
            Certificate: The created Certificate object.
        """

        ret = cls(
            version=cert.version.value,
            serial_number=cert.serial_number,
            signature_algorithm=SignatureAlgorithm.from_cryptography(
                cert.signature_hash_algorithm,
                cert.signature_algorithm_parameters,
            ),
            issuer=Name.from_cryptography(cert.issuer),
            validity=Validity(
                not_before=cert.not_valid_before_utc,
                not_after=cert.not_valid_after_utc,
            ),
            subject=Name.from_cryptography(cert.subject),
            subject_public_key_info=KeyPair.from_cryptography(
                cert.public_key()
            ),
            extensions=Extensions.from_cryptography(cert.extensions),
            signature_value=_byte_to_hex(cert.signature),
            _x509_obj=cert,
        )
        ret._x509_obj = cert
        return ret

    @classmethod
    def from_pem_string(
        cls: Type["Certificate"], cert_pem: str
    ) -> "Certificate":
        """
        Loads a certificate from a PEM string into a
        [Certificate][pki_tools.types.certificate.Certificate]
        object

        Arguments:
            cert_pem: The PEM encoded certificate in string format

        Returns:
            A Certificate created from the PEM

        Raises:
            CertLoadError: If the certificate could not be loaded
        """
        try:
            cert_pem = re.sub(r"\n\s*", "\n", cert_pem)
            if not _is_pem_cert_string(cert_pem):
                raise ValueError

            crypto_cert = x509.load_pem_x509_certificate(cert_pem.encode())
            return Certificate.from_cryptography(crypto_cert)
        except ValueError as e:
            logger.bind(cert=cert_pem).debug("Failed to load cert from PEM")
            raise CertLoadError(e)

    @classmethod
    def from_file(cls: Type["Certificate"], file_path: str) -> "Certificate":
        """
        Reads a file containing one PEM certificate into a
        [Certificate][pki_tools.types.certificate.Certificate]
        object

        Arguments:
            file_path:  Path and filename of the PEM certificate

        Returns:
             The Certificate loaded from the specified file
        """

        with open(file_path, "r") as f:
            cert_pem = f.read()

        return Certificate.from_pem_string(cert_pem)

    @classmethod
    def from_server(
        cls: Type["Certificate"],
        uri: str,
        cache_time_seconds: int = CACHE_TIME_SECONDS,
    ) -> "Certificate":
        """
        Loads a server certificate from a URI

        Args:
            uri: The https URI of the server containing the certificate
            cache_time_seconds: How long the request should be cached in memory

        Returns:
            The loaded [Certificate][pki_tools.types.certificate] object
        """
        cert_uri = CertsUri(uri=uri, cache_time_seconds=cache_time_seconds)

        cache_ttl = round(time.time() / cert_uri.cache_time_seconds)
        pem = _download_server_certificate(cert_uri.hostname, cache_ttl)
        return Certificate.from_pem_string(pem)

    @classmethod
    def from_uri(
        cls: Type["Certificate"],
        uri: str,
        cache_time_seconds: int = CACHE_TIME_SECONDS,
    ) -> "Certificate":
        """
        Loads Certificates from a URI.

        Args:
            uri: URI where the certificate can be downloaded.
            cache_time_seconds: Specifies how long the certificate
                should be cached, default is 1 month.
                Defaults to CACHE_TIME_SECONDS.

        Returns:
            Instance of Certificate containing the certificates
            fetched from the URI.
        """

        cache_ttl = round(time.time() / cache_time_seconds)
        cert_uri = CertsUri(uri=uri)
        pem = _download_pem(cert_uri.uri, cache_ttl)

        return Certificate.from_pem_string(pem)

    @property
    def tbs_bytes(self) -> bytes:
        """
        Returns:
            The to be signed bytes of this certificate
        """
        return self._crypto_object.tbs_certificate_bytes

    @property
    def pem_bytes(self) -> bytes:
        """
        Returns:
            Certificate PEM bytes
        """
        return self._crypto_object.public_bytes(serialization.Encoding.PEM)

    @property
    def pem_string(self) -> str:
        """
        Returns:
            Certificate PEM decoded into a string
        """
        return self.pem_bytes.decode()

    @property
    def hex_serial(self) -> str:
        """
        Parses the certificate serial into hex format

        Returns:
            String representing the hex value of the certificate serial number
        """
        hex_serial = format(self.serial_number, "x").zfill(32)
        return hex_serial.upper()

    @property
    def public_key(self) -> bytes:
        """
        Returns:
            The bytes of the public key in PEM format
        """
        return self._crypto_object.public_key().public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    @property
    def sign_alg_oid_name(self) -> str:
        """
        Returns:
            The name of the signature algorithm such as: `SHA512WITHRSA`
        """
        name = self._crypto_object.signature_algorithm_oid._name.upper()
        return name.replace("ENCRYPTION", "")

    @property
    def der_public_key(self) -> bytes:
        """
        Returns:
            The bytes of the public key in DER format
        """
        return _der_key(self._crypto_object.public_key())

    def digest(
        self,
        algorithm: HashAlgorithm = HashAlgorithm(
            name=HashAlgorithmName.SHA512
        ),
    ) -> str:
        """
        Gets the base64 encoded fingerprint of the certificate

        Args:
            algorithm: The algorithm to use to hash the fingerprint with

        Returns:
            Hashed and base64 encoded certificate fingerprint
        """
        fingerprint = self._crypto_object.fingerprint(
            algorithm._to_cryptography()
        )
        return base64.urlsafe_b64encode(fingerprint).decode("ascii")

    def to_file(self, file_path: str) -> None:
        """
        Saves the certificate PEM string to the specified file,
        creating it if it doesn't exist.

        Args:
            file_path: The path to the file (can be relative the caller or
                absolute)
        """
        with open(file_path, "w") as f:
            f.write(self.pem_string)

    def verify_signature(
        self: Type["Certificate"],
        signed: InitCryptoParser,
    ) -> None:
        """
        Verifies a signature of a signed entity against this issuer certificate

        Args:
            signed: The signed entity can either be a
                [Certificate][pki_tools.types.certificate.Certificate],
                [CertificateRevocationList][pki_tools.types.crl.CertificateRevocationList]
                or a [OCSPResponse][pki_tools.types.ocsp.OCSPResponse]

        Raises:
            InvalidSignedType: When the issuer has a non-supported type
            SignatureVerificationFailed: When the signature verification fails
        """
        try:
            self._crypto_object.public_key().verify(
                signed._crypto_object.signature,
                signed.tbs_bytes,
                PKCS1v15Padding()._to_cryptography(),
                signed._crypto_object.signature_hash_algorithm,
            )
            logger.trace("Signature valid")
        except Exception as e:
            logger.bind(
                exceptionType=type(e).__name__,
                exception=str(e),
            ).error("Signature verification failed")
            raise SignatureVerificationFailed(
                f"signature doesn't match issuer "
                f"with subject: {str(self.subject)}"
            )

    def sign(
        self,
        key_pair: CryptoKeyPair,
        signature_algorithm: SignatureAlgorithm,
        req_key: Optional[CryptoKeyPair] = None,
    ) -> None:
        """
        Signs a created [Certificate][pki_tools.types.certificate.Certificate]
        object with a given
        [CryptoKeyPair][pki_tools.types.key_pair.CryptoKeyPair]

        Args:
            key_pair: Keypair containing the private key to sing the
                certificate with
            signature_algorithm: Algorithm to use for the signature
            req_key: Can be used to sign another public key, defaults to the
                public key part in `key_pair`
        """
        self._private_key = key_pair
        self.serial_number = random.randint(1, 2**32 - 1)
        self.signature_algorithm = signature_algorithm

        if req_key is not None:
            self.subject_public_key_info = KeyPair(
                algorithm=req_key.__class__.__name__,
                parameters=req_key._string_dict(),
            )
            self.subject_public_key_info.create(req_key)

        self._x509_obj = self._to_cryptography()

    def __str__(self) -> str:
        return yaml.safe_dump(
            self._string_dict(),
            indent=2,
            default_flow_style=False,
            explicit_start=False,
            default_style="",
        )

    def _to_cryptography(self) -> x509.Certificate:
        if hasattr(self, "_x509_obj"):
            return self._x509_obj

        if not hasattr(self, "_private_key"):
            raise MissingInit(
                f"Please use Certificate.{self._init_func} " f"function"
            )

        subject = issuer = self.subject._to_cryptography()
        crypto_key = self._private_key._to_cryptography()
        if not hasattr(crypto_key, "public_key"):
            raise MissingInit("Invalid key type, use private key")

        public_key = crypto_key.public_key()
        if self.subject_public_key_info is not None:
            pub_key_pair = self.subject_public_key_info._key_pair
            public_key = pub_key_pair._to_cryptography().public_key()

        cert_builder = (
            x509.CertificateBuilder()
            .subject_name(
                subject,
            )
            .issuer_name(
                issuer,
            )
            .serial_number(
                x509.random_serial_number(),
            )
            .public_key(public_key)
            .not_valid_before(
                self.validity.not_before,
            )
            .not_valid_after(
                self.validity.not_after,
            )
        )

        if self.extensions is not None:
            for extension in self.extensions:
                cert_builder = cert_builder.add_extension(
                    extension._to_cryptography(), extension.critical
                )

        alg = self.signature_algorithm.algorithm._to_cryptography()
        cert = cert_builder.sign(crypto_key, alg)

        return cert

    def _string_dict(self):
        subject_key_info = self.subject_public_key_info._string_dict()
        signature_alg = self.signature_algorithm.algorithm.name.value
        return {
            "Version": self.version,
            "Serial Number": self.hex_serial,
            "Signature Algorithm": signature_alg,
            "Issuer": str(self.issuer),
            "Validity": self.validity._string_dict(),
            "Subject": str(self.subject),
            "Subject Public Key Info": subject_key_info,
            "Extensions": self.extensions._string_dict(),
            "Signature Value": self.signature_value,
        }


def _is_pem_cert_string(check: str):
    if not isinstance(check, str):
        return False

    return re.match(PEM_CERT_REGEX, check)


def _is_pem_csr_string(check: str):
    if not isinstance(check, str):
        return False

    return re.match(PEM_CSR_REGEX, check)



from datetime import datetime
from typing import Type, Optional, Dict, List

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from loguru import logger

from pki_tools.types.extensions import Extensions
from pki_tools.exceptions import CrlLoadError, MissingInit
from pki_tools.types.key_pair import CryptoKeyPair
from pki_tools.types.crypto_parser import CryptoParser, InitCryptoParser
from pki_tools.types.name import Name
from pki_tools.types.signature_algorithm import HashAlgorithm


class RevokedCertificate(CryptoParser):
    """
    Represents a revoked certificate.

    Attributes:
        serial: The serial number of the certificate.
        date: The revocation date of the certificate.
        extensions: Extensions associated with the certificate.
    """

    serial: int
    date: datetime
    extensions: Optional[Extensions] = None

    @classmethod
    def from_cryptography(
        cls: Type["RevokedCertificate"], crypto_obj: x509.RevokedCertificate
    ) -> "RevokedCertificate":
        """
        Create a RevokedCertificate object from a cryptography
        RevokedCertificate.

        Args:
            crypto_obj: Cryptography RevokedCertificate.

        Returns:
            RevokedCertificate: Instance of RevokedCertificate.
        """
        extensions = None
        if crypto_obj.extensions:
            extensions = Extensions.from_cryptography(crypto_obj.extensions)

        ret = cls(
            serial=crypto_obj.serial_number,
            date=crypto_obj.revocation_date,
            extensions=extensions,
            _x509_obj=crypto_obj,
        )

        return ret

    def _to_cryptography(self) -> x509.RevokedCertificate:
        builder = (
            x509.RevokedCertificateBuilder()
            .serial_number(self.serial)
            .revocation_date(self.date)
        )

        if self.extensions is not None:
            for extension in self.extensions:
                builder = builder.add_extension(
                    extval=extension._to_cryptography(),
                    critical=extension.critical,
                )

        return builder.build()

    def _string_dict(self) -> Dict:
        return {
            "Serial Number": self.serial,
            "Revocation Date": str(self.date),
            "Extensions": self.extensions._string_dict(),
        }


class CertificateRevocationList(InitCryptoParser):
    """
    Represents a certificate revocation list (CRL).

    Attributes:
        issuer: The name of the issuer.
        last_update: The last update time of the CRL.
        next_update: The next update time of the CRL.
        revoked_certs: List of revoked certificates.
    """

    issuer: Name
    last_update: datetime
    next_update: datetime

    revoked_certs: Optional[List[RevokedCertificate]] = None

    @classmethod
    def from_cryptography(
        cls: Type["CertificateRevocationList"],
        crypto_crl: x509.CertificateRevocationList,
    ) -> "CertificateRevocationList":
        """
        Create a CertificateRevocationList object from a cryptography
        CertificateRevocationList.

        Args:
            crypto_crl: Cryptography CertificateRevocationList.

        Returns:
            Instance of CertificateRevocationList.
        """
        ret = cls(
            issuer=Name.from_cryptography(crypto_crl.issuer),
            last_update=crypto_crl.last_update_utc,
            next_update=crypto_crl.next_update_utc,
        )
        ret._x509_obj = crypto_crl
        return ret

    @classmethod
    def from_bytes(
        cls: Type["CertificateRevocationList"], data: bytes
    ) -> "CertificateRevocationList":
        """
        Load a CertificateRevocationList object from bytes data.

        Args:
            data: Bytes data in DER or PEM format containing the CRL.

        Raises:
            CrlLoadError: If loading of CRL fails.

        Returns:
            Instance of CertificateRevocationList.
        """
        try:
            crypto_crl = x509.load_der_x509_crl(data)
            return CertificateRevocationList.from_cryptography(crypto_crl)
        except (TypeError, ValueError) as e:
            logger.bind(error=str(e)).trace("Error during loading of CRL DER")
            pass

        try:
            crypto_crl = x509.load_pem_x509_crl(data)
            return CertificateRevocationList.from_cryptography(crypto_crl)
        except TypeError as e:
            logger.bind(crl=data).error("Failed to load CRL")
            raise CrlLoadError(e) from None

    @property
    def tbs_bytes(self) -> bytes:
        """
        Return the bytes to be signed of the CRL.

        Returns:
            TBS bytes of the CRL.
        """
        return self._crypto_object.tbs_certlist_bytes

    @property
    def der_bytes(self) -> bytes:
        """
        Return the DER bytes of the CRL.

        Returns:
            DER bytes of the CRL.
        """
        return self._crypto_object.public_bytes(serialization.Encoding.DER)

    def sign(
        self, private_key: CryptoKeyPair, algorithm: HashAlgorithm
    ) -> None:
        """
        Sign the CRL with the provided private key and algorithm.

        Args:
            private_key: Key pair containing the private key used to
                sign the CRL.
            algorithm: Hash algorithm to use for signing.
        """
        self._private_key = private_key
        self._algorithm = algorithm
        self._x509_obj = self._to_cryptography()

    def get_revoked(self, cert_serial: int) -> Optional[RevokedCertificate]:
        """
        Get a revoked certificate by serial number.

        Args:
            cert_serial: Serial number of the certificate.

        Returns:
            RevokedCertificate object if found, else None.
        """
        crypto_revoked = (
            self._crypto_object.get_revoked_certificate_by_serial_number(
                cert_serial
            )
        )
        if crypto_revoked is not None:
            return RevokedCertificate.from_cryptography(crypto_revoked)
        return None

    def to_file(self, file_path: str) -> None:
        """
        Save the CRL to a file.

        Args:
            file_path: Path to save the file.
        """
        with open(file_path, "w") as f:
            f.write(self.der_bytes.decode())

    def _to_cryptography(self) -> x509.CertificateRevocationList:
        if not hasattr(self, "_private_key"):
            raise MissingInit("Please use 'sign' function")

        builder = x509.CertificateRevocationListBuilder()
        builder = builder.issuer_name(self.issuer._to_cryptography())
        builder = builder.last_update(self.last_update)
        builder = builder.next_update(self.next_update)

        for cert in self.revoked_certs:
            builder = builder.add_revoked_certificate(cert._to_cryptography())

        return builder.sign(
            private_key=self._private_key._to_cryptography(),
            algorithm=self._algorithm._to_cryptography(),
        )

    def _string_dict(self) -> Dict[str, str]:
        certs = []
        for cert in self.revoked_certs:
            certs.append(cert._string_dict())

        return {"Issuer": self.name._string_dict(), "Certs": certs}



from .types import (
    Chain,
    Certificate,
    Certificates,
    Extensions,
    Name,
    RevokeMode,
    SignatureAlgorithm,
    HashAlgorithm,
    HashAlgorithmName,
    Validity,
    CertificateSigningRequest,
    CertificateRevocationList,
    DSAKeyPair,
    RSAKeyPair,
    EllipticCurveKeyPair,
    Ed25519KeyPair,
    Ed448KeyPair,
    SHA1,
    SHA512_224,
    SHA512_256,
    SHA224,
    SHA256,
    SHA384,
    SHA512,
    SHA3_224,
    SHA3_256,
    SHA3_384,
    SHA3_512,
    SHAKE128,
    SHAKE256,
    MD5,
    BLAKE2b,
    BLAKE2s,
    SM3,
)

from .exceptions import (
    ExtensionMissing,
    Error,
    OcspInvalidResponseStatus,
    OcspFetchFailure,
    OcspIssuerFetchFailure,
    RevokeCheckFailed,
    NotCompleteChain,
    CertIssuerMissingInChain,
    CertExpired,
    InvalidSignedType,
    SignatureVerificationFailed,
    CertLoadError,
)

from .funcs import is_revoked, is_revoked_multiple_issuers



from enum import Enum


class RevokeMode(Enum):
    """
    Specifies in what precedence and protocols (OCSP, CRL)
    to check revocation status of a certificate

    Attributes:
        OCSP_FALLBACK_CRL: Check OCSP first and fallback to CRL if it fails
        OCSP_ONLY: Only check OCSP and ignore CRL
        CRL_ONLY: Only check CRL and ignore OCSP
    """

    OCSP_FALLBACK_CRL = "OCSP_FALLBACK_CRL"
    OCSP_ONLY = "OCSP_ONLY"
    CRL_ONLY = "CRL_ONLY"



import base64
import hashlib
from datetime import datetime
from enum import Enum
from typing import Type, Optional, Dict

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives._serialization import Encoding
from cryptography.x509 import ocsp
from cryptography.x509.ocsp import OCSPCertStatus
from loguru import logger

from pki_tools.types.extensions import Extensions
from pki_tools.types.certificate import Certificate
from pki_tools.exceptions import (
    MissingInit,
)
from pki_tools.types.key_pair import CryptoKeyPair
from pki_tools.types.crypto_parser import InitCryptoParser
from pki_tools.types.signature_algorithm import HashAlgorithm
from pki_tools.types.utils import _byte_to_hex


class OcspResponseStatus(Enum):
    """
    Enumeration of OCSP response statuses.
    """

    SUCCESSFUL = "SUCCESSFUL"
    MALFORMED_REQUEST = "MALFORMED_REQUEST"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    TRY_LATER = "TRY_LATER"
    SIG_REQUIRED = "SIG_REQUIRED"
    UNAUTHORIZED = "UNAUTHORIZED"


class OcspCertificateStatus(Enum):
    """
    Enumeration of OCSP certificate statuses.
    """

    GOOD = "GOOD"
    REVOKED = "REVOKED"
    UNKNOWN = "UNKNOWN"


class OCSPResponse(InitCryptoParser):
    """
    Represents an OCSP response.

    Attributes:
        response_status: The OCSP response status.
        certificate_status: The OCSP certificate status.
        issuer_key_hash: The issuer key hash.
        revocation_time: The revocation time.
    """

    response_status: OcspResponseStatus
    certificate_status: Optional[OcspCertificateStatus] = None
    issuer_key_hash: Optional[str] = None
    revocation_time: Optional[datetime] = None

    @classmethod
    def from_cryptography(
        cls: Type["OCSPResponse"], crypto_ocsp_response: ocsp.OCSPResponse
    ) -> "OCSPResponse":
        """
        Constructs an OCSPResponse object from a cryptography
        OCSPResponse object.

        Args:
            crypto_ocsp_response: The cryptography OCSPResponse object.

        Returns:
            OCSPResponse: The constructed OCSPResponse object.
        """
        response_status = crypto_ocsp_response.response_status

        ocsp_response_key_hash = None
        certificate_status = None
        revocation_time = None
        if response_status == ocsp.OCSPResponseStatus.SUCCESSFUL:
            certificate_status = crypto_ocsp_response.certificate_status.name
            try:
                ocsp_response_key_hash = _byte_to_hex(
                    crypto_ocsp_response.issuer_key_hash
                )
            except Exception as e:
                logger.bind(
                    exceptionType=type(e),
                    exception=str(e),
                    issuerHash=crypto_ocsp_response.issuer_key_hash,
                ).error("Couldn't convert issuer key hash to hex")
                raise

            if crypto_ocsp_response.revocation_time is not None:
                revocation_time = crypto_ocsp_response.revocation_time

        ret = cls(
            response_status=getattr(OcspResponseStatus, response_status.name),
            certificate_status=certificate_status,
            issuer_key_hash=ocsp_response_key_hash,
            revocation_time=revocation_time,
            _x509_obj=crypto_ocsp_response,
        )
        ret._x509_obj = crypto_ocsp_response
        return ret

    @classmethod
    def from_der_bytes(
        cls: Type["OCSPResponse"], der: bytes
    ) -> "OCSPResponse":
        """
        Constructs an OCSPResponse object from DER-encoded bytes.

        Args:
            der: The DER-encoded bytes.

        Returns:
            The constructed OCSPResponse object.
        """
        crypto_obj = ocsp.load_der_ocsp_response(der)
        return OCSPResponse.from_cryptography(crypto_obj)

    @property
    def tbs_bytes(self) -> bytes:
        """
        Returns the bytes to be singed of the OCSP response.

        Returns:
            bytes: The TBS bytes.
        """
        return self._crypto_object.tbs_response_bytes

    @property
    def der_bytes(self) -> bytes:
        """
        Returns the DER bytes of the OCSP response.

        Returns:
            bytes: The DER bytes.
        """
        return self._crypto_object.public_bytes(Encoding.DER)

    def hash_with_alg(self, der_key: bytes) -> str:
        """
        Hashes a DER key bytes with the algorithm of the OCSP response.

        Args:
            der_key: The DER key.

        Returns:
            str: The hashed key.
        """
        hash_algorithm = hashlib.new(self._crypto_object.hash_algorithm.name)
        hash_algorithm.update(der_key)
        return hash_algorithm.hexdigest().upper()

    @property
    def is_successful(self) -> bool:
        """
        Checks if the OCSP response is successful.

        Returns:
            True if the response is successful, False otherwise.
        """
        return self.response_status == OcspResponseStatus.SUCCESSFUL

    @property
    def is_revoked(self) -> bool:
        """
        Checks if the certificate is revoked.

        Returns:
            True if the certificate is revoked, False otherwise.
        """
        return self.certificate_status == OcspCertificateStatus.REVOKED

    def sign(
        self,
        cert: Certificate,
        issuer: Certificate,
        algorithm: HashAlgorithm,
        key_pair: CryptoKeyPair,
    ):
        """
        Signs the OCSP response.

        Args:
            cert: The certificate.
            issuer: The issuer certificate.
            algorithm: The hash algorithm.
            key_pair: The key pair.
        """
        self._cert = cert
        self._issuer = issuer
        self._algorithm = algorithm
        self._private_key = key_pair
        self._x509_obj = self._to_cryptography()

    def _to_cryptography(self) -> ocsp.OCSPResponse:
        if not hasattr(self, "_private_key"):
            raise MissingInit(
                f"Please use OCSPResponse.{self._init_func} function"
            )

        builder = ocsp.OCSPResponseBuilder()
        cert_status = None
        if self.certificate_status is not None:
            cert_status = getattr(OCSPCertStatus, self.certificate_status.name)

        builder = builder.add_response(
            cert=self._cert._to_cryptography(),
            issuer=self._issuer._to_cryptography(),
            algorithm=self._algorithm._to_cryptography(),
            cert_status=cert_status,
            this_update=datetime.now(),
            next_update=datetime.now(),
            revocation_reason=None,
            revocation_time=self.revocation_time,
        ).responder_id(
            ocsp.OCSPResponderEncoding.HASH, self._cert._to_cryptography()
        )

        return builder.sign(
            self._private_key._to_cryptography(),
            self._algorithm._to_cryptography(),
        )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "Response Status": self.response_status,
            "Certificate Status": self.certificate_status,
            "Issuer Key Hash": self.issuer_key_hash,
        }


class OCSPRequest(InitCryptoParser):
    """
    Represents an OCSP request.

    Attributes:
        hash_algorithm: The hash algorithm.
        serial_number: The serial number.
        extensions: The extensions.
    """

    hash_algorithm: HashAlgorithm

    serial_number: Optional[int] = None
    extensions: Optional[Extensions] = None

    _init_func = "create"

    def from_cryptography(
        cls: Type["OCSPRequest"], crypto_obj: ocsp.OCSPRequest
    ) -> "OCSPRequest":
        """
        Constructs an OCSPRequest object from a cryptography OCSPRequest
        object.

        Args:
            crypto_obj: The cryptography OCSPRequest object.

        Returns:
            The constructed OCSPRequest object.
        """
        return cls(
            serial_number=crypto_obj.serial_number,
            extensions=Extensions.from_cryptography(crypto_obj.extensions),
            _x590_obj=crypto_obj,
        )

    @property
    def request_path(self) -> str:
        """
        The request path of the OCSP Response.

        Returns:
            The request path.
        """
        return base64.b64encode(
            self._crypto_object.public_bytes(serialization.Encoding.DER)
        ).decode()

    def create(self, cert: Certificate, issuer_cert: Certificate):
        """
        Creates an OCSP request.

        Args:
            cert: The certificate.
            issuer_cert: The issuer of the OCSP Response.
        """
        self._cert = cert
        self._issuer = issuer_cert
        self._x509_obj = self._to_cryptography()

    def _to_cryptography(self) -> ocsp.OCSPRequest:
        if not hasattr(self, "_cert"):
            raise MissingInit(
                f"Please use " f"OCSPRequest.{self._init_func} function"
            )

        builder = ocsp.OCSPRequestBuilder()
        builder = builder.add_certificate(
            self._cert._to_cryptography(),
            self._issuer._to_cryptography(),
            self.hash_algorithm._to_cryptography(),
        )
        return builder.build()

    def _string_dict(self) -> Dict:
        ret = self.hash_algorithm._string_dict()

        if self.serial_number is not None:
            ret["Serial Number"] = self.serial_number
        if self.extensions is not None:
            ret["Extensions"] = self.extensions._string_dict()

        return ret



import time
from typing import List, Type

from cryptography import x509
from loguru import logger

from pki_tools.types.certificate import Certificate
from pki_tools.types.crypto_parser import CryptoParser
from pki_tools.types.utils import CACHE_TIME_SECONDS


class Certificates(CryptoParser):
    """
    A list of one or more certificates

    Attributes:
        certificates: a list of
            [Certificate][pki_tools.types.certificate.Certificate]
    """

    certificates: List[Certificate]

    @classmethod
    def from_cryptography(
        cls: Type["Certificates"], crypto_certs: List[x509.Certificate]
    ) -> "Certificates":
        """
        Create a Certificates object from a list of cryptography certificates.

        Args:
            crypto_certs: List of cryptography certificates.

        Returns:
            Instance of Certificates containing the provided certificates.
        """
        certificates = []
        for crypt_cert in crypto_certs:
            certificates.append(Certificate.from_cryptography(crypt_cert))
        return cls(certificates=certificates, _x509_obj=crypto_certs)

    @classmethod
    def from_file(cls: Type["Certificates"], file_path: str) -> "Certificates":
        """
        Reads a file containing one or more PEM certificate(s) into a
        Certificates object.

        Args:
            file_path: Path and filename of the PEM certificate.

        Returns:
            A Certificates object representing the certificate(s) from file.
        """
        with open(file_path, "r") as f:
            cert_pem = f.read()

        crypto_certs = x509.load_pem_x509_certificates(cert_pem.encode())
        return cls.from_cryptography(crypto_certs)

    @classmethod
    def from_pem_string(
        cls: Type["Certificates"], pem_string: str
    ) -> "Certificates":
        """
        Create a Certificates object from a PEM string.

        Args:
            pem_string: PEM string containing certificate(s).

        Returns:
            Instance of Certificates containing the certificates from the PEM
            string.
        """
        crypto_certs = x509.load_pem_x509_certificates(pem_string.encode())
        return cls.from_cryptography(crypto_certs)

    @classmethod
    def from_uri(
        cls: Type["Certificates"],
        uris: [str],
        cache_time_seconds: int = CACHE_TIME_SECONDS,
    ) -> "Certificates":
        """
        Loads Certificates from one or more URI(s).

        Args:
            uris: One or more URI(s) where the certificate(s) can
                be downloaded.
            cache_time_seconds: Specifies how long the certificates
                should be cached, default is 1 month.
                Defaults to CACHE_TIME_SECONDS.

        Returns:
            Instance of Certificates containing the certificates
            fetched from the URI.
        """

        certificates = []
        cache_ttl = round(time.time() / cache_time_seconds)
        for uri in uris:
            certificates.append(Certificate.from_uri(uri, cache_ttl))

        return cls(certificates=certificates)

    @property
    def pem_string(self) -> str:
        """
        Returns a string containing the PEM-encoded certificates.

        Returns:
            PEM string containing all the certificates.
        """
        all_certs = ""
        for cert in self.certificates:
            all_certs += cert.pem_string

        return all_certs

    def to_file(self, file_path: str) -> None:
        """
        Saves one or more certificate(s) into a file.

        Args:
            file_path: Path and filename where to store the certificate(s).
        """
        with open(file_path, "w") as f:
            f.write(self.pem_string)

        logger.debug(f"Certificate(s) saved to {file_path}")

    def _to_cryptography(self) -> List[x509.Certificate]:
        certs = []
        for cert in self.certificates:
            certs.append(cert._to_cryptography())
        return certs

    def _string_dict(self):
        certs = {"Certificates": []}
        for cert in self.certificates:
            certs["Certificates"].append(cert._string_dict())
        return certs

    def __str__(self):
        ret = ""
        count = 0
        for cert in self.certificates:
            count += 1
            ret += f"{'-'*10}Certificate #{count}{'-'*10}\n"
            ret += str(cert)
            ret += f"{'-'*34}\n"
        ret += f"Chain certificate count: {count}"
        return ret



import importlib
from abc import ABC
from enum import Enum
from typing import Type, Optional, Union, Dict

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives._serialization import Encoding
from cryptography.hazmat.primitives.asymmetric import padding, ec
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from pydantic import BaseModel, ConfigDict

from pki_tools.exceptions import MissingBlockSize
from pki_tools.types.crypto_parser import CryptoParser

HASHES_MODULE = importlib.import_module(
    "cryptography.hazmat.primitives.hashes"
)


class HashAlgorithmName(Enum):
    """
    Enumeration of hash algorithm names.
    """

    SHA1 = "SHA1"
    SHA512_224 = "SHA512_224"
    SHA512_256 = "SHA512_256"
    SHA224 = "SHA224"
    SHA256 = "SHA256"
    SHA384 = "SHA384"
    SHA512 = "SHA512"
    SHA3_224 = "SHA3_224"
    SHA3_256 = "SHA3_256"
    SHA3_384 = "SHA3_384"
    SHA3_512 = "SHA3_512"
    SHAKE128 = "SHAKE128"
    SHAKE256 = "SHAKE256"
    MD5 = "MD5"
    BLAKE2b = "BLAKE2b"
    BLAKE2s = "BLAKE2s"
    SM3 = "SM3"


class HashAlgorithm(CryptoParser):
    """
    Represents a hash algorithm.

    Attributes:
        name: The name of the hash algorithm.
        block_size: The block size of the hash algorithm.
    """

    name: HashAlgorithmName
    block_size: Optional[int] = None

    @classmethod
    def from_cryptography(
        cls: Type["HashAlgorithm"], x509_obj: hashes.HashAlgorithm
    ) -> "HashAlgorithm":
        """
        Constructs a HashAlgorithm object from a cryptography HashAlgorithm
        object.

        Args:
            x509_obj: The cryptography HashAlgorithm object.

        Returns:
            The constructed HashAlgorithm object.
        """
        return cls(
            name=x509_obj.__class__.__name__,
            block_size=x509_obj.block_size,
            _x509_obj=x509_obj,
        )

    @property
    def der_bytes(self):
        return self._to_cryptography().public_bytes(Encoding.DER)

    def _to_cryptography(self) -> hashes.HashAlgorithm:
        if "SHAKE" in self.name.value:
            if self.block_size is None:
                raise MissingBlockSize("Please set block_size")

            return getattr(HASHES_MODULE, self.name.value)(self.block_size)

        if self.name.value == "BLAKE2s":
            return getattr(HASHES_MODULE, self.name.value)(32)
        elif self.name.value == "BLAKE2b":
            return getattr(HASHES_MODULE, self.name.value)(64)

        return getattr(HASHES_MODULE, self.name.value)()

    def _string_dict(self) -> Dict[str, str]:
        return {"algorithm": self.name.value}


class Padding(CryptoParser, ABC):
    """
    Abstract base class representing padding.
    """


class PSSPaddingLenght(Enum):
    """
    Enumeration of PSS padding lengths.
    """

    _MaxLength = "_MaxLength"
    _Auto = "_Auto"
    _DigestLength = "_DigestLength"


class PSSPadding(Padding):
    """
    Represents PSS padding.

    Attributes:
        mgf: The mask generation function.
        length: The length of the padding.
    """

    mgf: HashAlgorithm
    length: Union[int, PSSPaddingLenght]

    @classmethod
    def from_cryptography(
        cls: Type["PSSPadding"], crypto_obj: padding.PSS
    ) -> "PSSPadding":
        """
        Constructs a PSSPadding object from a cryptography PSS object.

        Args:
            crypto_obj: The cryptography PSS object.

        Returns:
            The constructed PSSPadding object.
        """
        salt_length = crypto_obj._salt_length
        if not isinstance(salt_length, int):
            salt_length = PSSPaddingLenght[salt_length.__class__.__name__]

        return cls(
            mgf=crypto_obj._mgf._algorithm,
            length=salt_length,
            _x509_obj=crypto_obj,
        )

    def _to_cryptography(self) -> padding.PSS:
        return padding.PSS(
            mgf=padding.MGF1(self.mgf._to_cryptography()),
            salt_length=getattr(padding, self.length.value),
        )

    def _string_dict(self) -> Dict:
        length = self.length
        if not isinstance(length, int):
            length = self.length.name

        return {
            "mgf": self.mgf._string_dict(),
            "salt_length": length,
        }


class PKCS1v15Padding(Padding):
    """
    Represents PKCS1v15 padding.
    """

    _name: str

    @classmethod
    def from_cryptography(
        cls: Type["PKCS1v15Padding"], crypto_obj: padding.PKCS1v15
    ) -> "PKCS1v15Padding":
        """
        Constructs a PKCS1v15Padding object from a cryptography PKCS1v15
        object.

        Args:
            crypto_obj: The cryptography PKCS1v15 object.

        Returns:
            The constructed PKCS1v15Padding object.
        """
        return cls(_name=crypto_obj.name, _x509_obj=crypto_obj)

    def _to_cryptography(self) -> padding.PKCS1v15:
        return padding.PKCS1v15()

    def _string_dict(self) -> Dict[str, str]:
        return {"name": self._name}


class ECDSAPadding(Padding):
    """
    Represents ECDSA padding.

    Attributes:
        algorithm: The hash algorithm.
        prehashed: Indicates if the data is prehashed.
    """

    algorithm: HashAlgorithm
    prehashed: bool

    @classmethod
    def from_cryptography(
        cls: Type["ECDSAPadding"], crypto_obj: ec.ECDSA
    ) -> "ECDSAPadding":
        """
        Constructs an ECDSAPadding object from a cryptography ECDSA object.

        Args:
            crypto_obj: The cryptography ECDSA object.

        Returns:
            The constructed ECDSAPadding object.
        """
        if isinstance(crypto_obj.algorithm, Prehashed):
            prehashed = True
            algorithm = crypto_obj.algorithm._algorithm
        else:
            prehashed = False
            algorithm = crypto_obj.algorithm

        algorithm = HashAlgorithm.from_cryptography(algorithm)

        return cls(
            algorithm=algorithm,
            prehashed=prehashed,
            _x509_obj=crypto_obj,
        )

    def _to_cryptography(self) -> ec.ECDSA:
        alg = self.algorithm._to_cryptography()
        if self.prehashed:
            alg = Prehashed(alg)

        return ec.ECDSA(algorithm=alg)

    def _string_dict(self) -> Dict:
        length = self.length
        if not isinstance(length, int):
            length = self.length.name

        ret = self.algorithm._string_dict()
        ret["prehashed"] = str(self.prehashed)
        return ret


class SignatureAlgorithm(BaseModel):
    """
    Represents a signature algorithm.

    Attributes:
        algorithm: The hash algorithm.
        parameters: The parameters of the signature algorithm.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    algorithm: HashAlgorithm
    parameters: Optional[
        Union[PSSPadding, PKCS1v15Padding, ECDSAPadding]
    ] = None

    @classmethod
    def from_cryptography(
        cls: Type["SignatureAlgorithm"],
        algorithm: hashes.HashAlgorithm,
        parameters: Union[padding.PSS, padding.PKCS1v15, ec.ECDSA] = None,
    ) -> "SignatureAlgorithm":
        """
        Constructs a SignatureAlgorithm object from cryptography objects.

        Args:
            algorithm: The hash algorithm.
            parameters: The parameters.

        Returns:
            The constructed SignatureAlgorithm object.
        """
        algorithm = HashAlgorithm.from_cryptography(algorithm)

        if parameters is None:
            return cls(algorithm=algorithm)

        if isinstance(parameters, padding.PSS):
            parameters = PSSPadding.from_cryptography(parameters)
        elif isinstance(parameters, padding.PKCS1v15):
            parameters = PKCS1v15Padding.from_cryptography(parameters)
        elif isinstance(parameters, ec.ECDSA):
            parameters = ECDSAPadding.from_cryptography(parameters)

        return cls(algorithm=algorithm, parameters=parameters)

    def _string_dict(self):
        ret = self.algorithm._string_dict()

        if self.parameters is not None:
            ret["parameters"] = self.parameters._string_dict()

        return ret


SHA1 = SignatureAlgorithm(algorithm=HashAlgorithm(name=HashAlgorithmName.SHA1))
SHA512_224 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA512_224)
)
SHA512_256 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA512_256)
)
SHA224 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA224)
)
SHA256 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA256)
)
SHA384 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA384)
)
SHA512 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA512)
)
SHA3_224 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA3_224)
)
SHA3_256 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA3_256)
)
SHA3_384 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA3_384)
)
SHA3_512 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHA3_512)
)
SHAKE128 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHAKE128, block_size=64)
)
SHAKE256 = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.SHAKE256, block_size=64)
)
MD5 = SignatureAlgorithm(algorithm=HashAlgorithm(name=HashAlgorithmName.MD5))
BLAKE2b = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.BLAKE2b)
)
BLAKE2s = SignatureAlgorithm(
    algorithm=HashAlgorithm(name=HashAlgorithmName.BLAKE2s)
)
SM3 = SignatureAlgorithm(algorithm=HashAlgorithm(name=HashAlgorithmName.SM3))



from collections import defaultdict
from typing import List, Type

from cryptography import x509
from cryptography.hazmat._oid import NameOID
from cryptography.hazmat.bindings._rust import ObjectIdentifier

from pydantic import Field, ConfigDict

from pki_tools.types.crypto_parser import CryptoParser


class Name(CryptoParser):
    """
    Name type describes e.g. certificate subject or issuer.
    The attributes are following the
    [RFC5280#Section-4.1.2.4](https://datatracker.ietf.org/doc/html/rfc5280#section-4.1.2.4)

    Note that every attribute is a list of string in order to support
    multivalued RDNs.

    Attributes:
        c: Country Name (2.5.4.6)
        o:  Organization Name (2.5.4.10)
        ou:  Organizational Unit Name (2.5.4.11)
        dnq:  Distinguished Name Qualifier (2.5.4.46)
        s:  State Or Province Name (2.5.4.8)
        cn:  Common Name (2.5.4.3)
        serial:  Serial Number (2.5.4.5)
        ln:  Locality Name (2.5.4.7)
        t:  Title (2.5.4.12)
        sn:  Surname (2.5.4.4)
        gn:  Given Name (2.5.4.42)
        i:  Initials (2.5.4.43)
        p:  Pseudonym (2.5.4.65)
        gq:  Generation Qualifier (2.5.4.44)
        dc:  Domain Component (0.9.2342.19200300.100.1.25)
    """

    model_config = ConfigDict(populate_by_name=True)

    c: List[str] = Field(alias=NameOID.COUNTRY_NAME.dotted_string, default=[])
    o: List[str] = Field(
        alias=NameOID.ORGANIZATION_NAME.dotted_string, default=[]
    )
    ou: List[str] = Field(
        alias=NameOID.ORGANIZATIONAL_UNIT_NAME.dotted_string, default=[]
    )
    dnq: List[str] = Field(
        alias=NameOID.DN_QUALIFIER.dotted_string, default=[]
    )
    s: List[str] = Field(
        alias=NameOID.STATE_OR_PROVINCE_NAME.dotted_string, default=[]
    )
    cn: List[str] = Field(alias=NameOID.COMMON_NAME.dotted_string, default=[])
    serial: List[str] = Field(
        alias=NameOID.SERIAL_NUMBER.dotted_string, default=[]
    )

    ln: List[str] = Field(
        alias=NameOID.LOCALITY_NAME.dotted_string, default=[]
    )
    t: List[str] = Field(alias=NameOID.TITLE.dotted_string, default=[])
    sn: List[str] = Field(alias=NameOID.SURNAME.dotted_string, default=[])
    gn: List[str] = Field(alias=NameOID.GIVEN_NAME.dotted_string, default=[])
    i: List[str] = Field(alias=NameOID.INITIALS.dotted_string, default=[])
    p: List[str] = Field(alias=NameOID.PSEUDONYM.dotted_string, default=[])
    gq: List[str] = Field(
        alias=NameOID.GENERATION_QUALIFIER.dotted_string, default=[]
    )
    dc: List[str] = Field(
        alias=NameOID.DOMAIN_COMPONENT.dotted_string, default=[]
    )

    @classmethod
    def from_cryptography(cls: Type["Name"], name: x509.Name) -> "Name":
        """
        Create a Name instance from a cryptography Name object.

        Args:
            name: The cryptography Name object.

        Returns:
            The Name instance.
        """
        subject = defaultdict(set)
        for attribute in name:
            for att in name.get_attributes_for_oid(attribute.oid):
                subject[att.oid.dotted_string].add(att.value)
        subject = dict(subject)
        subject["_x509_obj"] = name
        return cls(**subject)

    def _to_cryptography(self) -> x509.Name:
        name_attributes = []
        for name, field in self.model_fields.items():
            object_identifier = ObjectIdentifier(field.alias)
            field_vals = getattr(self, name)
            for val in field_vals:
                name_attr = x509.NameAttribute(object_identifier, val)
                name_attributes.append(name_attr)
        return x509.Name(name_attributes)

    def _string_dict(self):
        ret = defaultdict(list)
        for a in set(self.model_dump()):
            for val in getattr(self, a):
                ret[a.upper()].append(val)
        return ret

    def __str__(self):
        name_list = []
        for k, v in self._string_dict().items():
            name_list.append(f"{k}: {','.join(v)}")
        return ", ".join(name_list)

    def __eq__(self, other):
        for key in self.model_dump():
            val_list = getattr(self, key)
            if set(val_list) != set(getattr(other, key)):
                return False
        return True



class Error(Exception):
    pass


class ExtensionMissing(Error):
    pass


class CertLoadError(Error):
    """
    Thrown when the cert could not be loaded e.g. due to invalid PEM format
    """

    pass


class OcspFetchFailure(Error):
    pass


class FetchFailure(Error):
    """
    FetchFailure
    """


class OcspInvalidResponseStatus(Error):
    pass


class OcspIssuerFetchFailure(Error):
    pass


class CrlFetchFailure(Error):
    pass


class CrlLoadError(Error):
    """
    Thrown when CRL could not be loaded due to e.g. invalid format
    """

    pass


class CsrLoadError(Error):
    """
    CsrLoadError
    """

    pass


class RevokeCheckFailed(Error):
    """
    RevokeCheckFailed
    """

    pass


class CertIssuerMissingInChain(Error):
    """
    CertIssuerMissingInChain
    """

    pass


class NotCompleteChain(Error):
    """
    NotCompleteChain
    """

    pass


class CertExpired(Error):
    """
    CertExpired
    """

    pass


class SignatureVerificationFailed(Error):
    """
    SignatureVerificationFailed
    """

    pass


class InvalidSignedType(Error):
    """
    InvalidSignedType
    """

    pass


class MissingInit(Error):
    """
    Risen when the
    [InitCryptoParser][pki_tools.types.crypto_parser.InitCryptoParser]
    has not been initialized.
    """

    pass


class MissingBlockSize(Error):
    """
    MissingBlockSize
    """

    pass


class InvalidKeyType(Error):
    """
    MissingBlockSize
    """

    pass



import abc
import importlib
from typing import Dict, Type, Union, Optional, get_args

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import (
    dsa,
    ec,
    rsa,
    ed25519,
    ed448,
)
from cryptography.hazmat.primitives.asymmetric.types import (
    CertificateIssuerPrivateKeyTypes,
    CertificateIssuerPublicKeyTypes,
)

from pydantic import ConfigDict

from pki_tools.exceptions import InvalidKeyType
from pki_tools.types.crypto_parser import (
    CryptoObject,
    InitCryptoParser,
)
from pki_tools.types.utils import _byte_to_hex, _hex_to_byte, _der_key


class CryptoKeyPair(InitCryptoParser):
    """
    Represents a cryptographic key pair.
    """

    _init_func = "generate"

    @classmethod
    @abc.abstractmethod
    def generate(cls: Type["CryptoKeyPair"], *args) -> "CryptoKeyPair":
        """
        Abstract method to generate a cryptographic key pair.

        Returns:
            The generated cryptographic key pair.
        """
        raise NotImplementedError

    @property
    def der_public_key(self) -> bytes:
        """
        Property to get the DER encoding of the public key.

        Returns:
            bytes: The DER encoded public key.
        """
        return _der_key(self._crypto_object.public_key())

    @property
    def pem_public_key(self) -> bytes:
        """
        Property to get the PEM encoding of the public key.

        Returns:
            bytes: The PEM encoded public key.
        """
        public_key = self._crypto_object
        if hasattr(self._crypto_object, "public_key"):
            public_key = public_key.public_key()

        return public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    @property
    def pem_private_key(self) -> bytes:
        """
        Property to get the PEM encoding of the private key.

        Returns:
            bytes: The PEM encoded private key.
        """
        if not hasattr(self._crypto_object, "public_key"):
            raise InvalidKeyType("Can't get private key using public key")

        kwargs = {
            "encoding": serialization.Encoding.PEM,
            "encryption_algorithm": serialization.NoEncryption(),
            "format": serialization.PrivateFormat.TraditionalOpenSSL,
        }

        if "Ed" in self.__class__.__name__:
            kwargs["format"] = serialization.PrivateFormat.PKCS8

        return self._crypto_object.private_bytes(**kwargs)

    def private_key_to_file(self, file_path: str) -> None:
        """
        Method to write the PEM-encoded private key to a file.

        Args:
            file_path (str): Path to the output file.
        """
        with open(file_path, "w") as f:
            f.write(self.pem_private_key.decode())

    def public_key_to_file(self, file_path: str):
        """
        Method to write the PEM-encoded public key to a file.

        Args:
            file_path (str): Path to the output file.
        """
        with open(file_path, "w") as f:
            f.write(self.pem_public_key.decode())


class DSAKeyPair(CryptoKeyPair):
    """
    Represents a DSA cryptographic key pair.
    """

    key_size: int

    y: int
    p: int
    q: int
    g: int

    x: Optional[int]

    @classmethod
    def generate(cls: Type["DSAKeyPair"], key_size: int) -> "DSAKeyPair":
        """
        Generate a DSA cryptographic key pair.

        Args:
            key_size: The size of the key.

        Returns:
            The generated DSA key pair.
        """
        new_key = dsa.generate_private_key(key_size=key_size)
        return DSAKeyPair.from_cryptography(new_key)

    @classmethod
    def from_cryptography(
        cls: Type["DSAKeyPair"],
        key: Union[dsa.DSAPrivateKey, dsa.DSAPublicKey],
    ) -> "DSAKeyPair":
        """
        Create a DSAKeyPair from a cryptography key.

        Args:
            key: The [cryptography.hazmat.primitives.asymmetric.dsa][]
                public or private key.

        Returns:
            DSAKeyPair: The DSA key pair.
        """
        if isinstance(key, dsa.DSAPrivateKey):
            public_key = key.public_key()
            x = key.private_numbers().x
        elif isinstance(key, dsa.DSAPublicKey):
            public_key = key
            x = None
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

        public_numbers = public_key.public_numbers()
        param_numbers = public_key.parameters().parameter_numbers()

        ret = cls(
            x=x,
            y=public_numbers.y,
            p=param_numbers.p,
            q=param_numbers.q,
            g=param_numbers.g,
            key_size=key.key_size,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(self) -> Union[dsa.DSAPrivateKey, dsa.DSAPublicKey]:
        public_numbers = dsa.DSAPublicNumbers(
            y=self.y,
            parameter_numbers=dsa.DSAParameterNumbers(
                p=self.p, q=self.q, g=self.g
            ),
        )
        if self.x is None:
            return public_numbers.public_key()

        private_numbers = dsa.DSAPrivateNumbers(
            x=self.x, public_numbers=public_numbers
        )
        return private_numbers.private_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "key_size": str(self.key_size),
            "public_key_y": self.y,
            "prime_p": self.p,
            "subprime_q": self.q,
            "generator_g": self.g,
        }


class RSAKeyPair(CryptoKeyPair):
    """
    Represents an RSA cryptographic key pair.
    """

    key_size: int

    e: int
    n: int

    d: Optional[int]
    p: Optional[int]
    q: Optional[int]
    dmp1: Optional[int]
    dmq1: Optional[int]
    iqmp: Optional[int]

    @classmethod
    def generate(
        cls: Type["RSAKeyPair"], key_size: int = 2048, exponent: int = 65537
    ) -> "RSAKeyPair":
        """
        Generate an RSA cryptographic key pair.

        Args:
            key_size: The size of the key. Defaults to `2048`.
            exponent: The public exponent. Defaults to `65537`.

        Returns:
            The generated RSA key pair.
        """
        new_key = rsa.generate_private_key(
            public_exponent=exponent,
            key_size=key_size,
        )
        return RSAKeyPair.from_cryptography(new_key)

    @classmethod
    def from_cryptography(
        cls: Type["RSAKeyPair"],
        key: Union[rsa.RSAPrivateKey, rsa.RSAPublicKey],
    ) -> "RSAKeyPair":
        """
        Create an RSAKeyPair from a cryptography key.

        Args:
            key: The cryptography key.

        Returns:
            The RSA key pair.
        """
        if isinstance(key, rsa.RSAPrivateKey):
            private_numbers = key.private_numbers()
            public_numbers = private_numbers.public_numbers
            d = private_numbers.d
            p = private_numbers.p
            q = private_numbers.q
            dmp1 = private_numbers.dmp1
            dmq1 = private_numbers.dmq1
            iqmp = private_numbers.iqmp
        elif isinstance(key, rsa.RSAPublicKey):
            public_numbers = key.public_numbers()
            d = None
            p = None
            q = None
            dmp1 = None
            dmq1 = None
            iqmp = None
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

        ret = cls(
            e=public_numbers.e,
            n=public_numbers.n,
            d=d,
            p=p,
            q=q,
            dmp1=dmp1,
            dmq1=dmq1,
            iqmp=iqmp,
            key_size=key.key_size,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(self) -> Union[rsa.RSAPrivateKey, rsa.RSAPublicKey]:
        public_numbers = rsa.RSAPublicNumbers(e=self.e, n=self.n)

        if self.p is None:
            return public_numbers.public_key()

        private_numbers = rsa.RSAPrivateNumbers(
            p=self.p,
            q=self.q,
            d=self.d,
            dmp1=self.dmp1,
            dmq1=self.dmq1,
            iqmp=self.iqmp,
            public_numbers=public_numbers,
        )
        return private_numbers.private_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "key_size": str(self.key_size),
            "public_exponent_e": str(self.e),
            "modulus_n": str(self.n),
            "private_exponent_d": str(self.d),
            "prime_p": str(self.p),
            "prime_q": str(self.q),
            "dmp1": str(self.dmp1),
            "dmq1": str(self.dmq1),
            "iqmp": str(self.iqmp),
        }


EC_MODULE = importlib.import_module(
    "cryptography.hazmat.primitives.asymmetric.ec"
)


class EllipticCurveKeyPair(CryptoKeyPair):
    """
    Represents an elliptic curve cryptographic key pair.
    """

    curve_name: str

    x: int
    y: int

    d: Optional[int]

    @classmethod
    def generate(
        cls: Type["EllipticCurveKeyPair"], curve_name: str
    ) -> "EllipticCurveKeyPair":
        """
        Generate an elliptic curve cryptographic key pair.

        Args:
            curve_name: The name of the curve.

        Returns:
            The generated elliptic curve key pair.
        """
        allowed = [val.name.upper() for val in ec._CURVE_TYPES.values()]
        if curve_name.upper() not in allowed:
            raise TypeError(
                f"Curve Name: {curve_name} "
                f"doesn't match supported names: {allowed}"
            )

        new_key = ec.generate_private_key(
            curve=getattr(EC_MODULE, curve_name.upper())()
        )

        return EllipticCurveKeyPair.from_cryptography(new_key)

    @classmethod
    def from_cryptography(
        cls: Type["EllipticCurveKeyPair"],
        key: Union[ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey],
    ) -> "EllipticCurveKeyPair":
        """
        Create an EllipticCurveKeyPair from a cryptography key.

        Args:
            key: The cryptography key.

        Returns:
            The elliptic curve key pair.
        """
        if isinstance(key, ec.EllipticCurvePrivateKey):
            private_numbers = key.private_numbers()
            public_key = key.public_key()
            public_numbers = public_key.public_numbers()
            curve_name = public_key.curve.name.upper()
            d = private_numbers.private_value
        elif isinstance(key, ec.EllipticCurvePublicKey):
            public_numbers = key.public_numbers()
            curve_name = key.curve.name.upper()
            d = None
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

        ret = cls(
            x=public_numbers.x,
            y=public_numbers.y,
            curve_name=curve_name,
            d=d,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(
        self,
    ) -> Union[ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey]:
        curve = getattr(EC_MODULE, self.curve_name)()
        public_numbers = ec.EllipticCurvePublicNumbers(self.x, self.y, curve)
        if self.d is None:
            return public_numbers.public_key()

        private_numbers = ec.EllipticCurvePrivateNumbers(
            private_value=self.d, public_numbers=public_numbers
        )

        return private_numbers.private_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "curve_name": self.curve_name,
            "x_coordinate": str(self.x),
            "y_coordinate": str(self.y),
            "private_key_d": str(self.d),
        }


class Ed25519KeyPair(CryptoKeyPair):
    """
    Represents an Ed25519 cryptographic key pair.
    """

    public_bytes: str
    private_bytes: Optional[str]

    @classmethod
    def generate(cls: Type["Ed25519KeyPair"]) -> "Ed25519KeyPair":
        """
        Generate an Ed25519 cryptographic key pair.

        Returns:
            The generated Ed25519 key pair.
        """
        new_key = ed25519.Ed25519PrivateKey.generate()
        return Ed25519KeyPair.from_cryptography(new_key)

    @classmethod
    def from_cryptography(
        cls: Type["Ed25519KeyPair"],
        key: Union[ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey],
    ) -> "Ed25519KeyPair":
        """
        Create an Ed25519KeyPair from a cryptography key.

        Args:
            key: The cryptography key.

        Returns:
            The Ed25519 key pair.
        """
        private_bytes = None

        if isinstance(key, ed25519.Ed25519PrivateKey):
            public_key = key.public_key()
            private_bytes = _byte_to_hex(key.private_bytes_raw())
        elif isinstance(key, ed25519.Ed25519PublicKey):
            public_key = key
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

        ret = cls(
            public_bytes=_byte_to_hex(public_key.public_bytes_raw()),
            private_bytes=private_bytes,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(
        self,
    ) -> Union[ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey]:
        if self.private_bytes is not None:
            return ed25519.Ed25519PrivateKey.from_private_bytes(
                _hex_to_byte(self.private_bytes)
            )
        else:
            return ed25519.Ed25519PublicKey.from_public_bytes(
                _hex_to_byte(self.public_bytes)
            )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "public_bytes": self.public_bytes,
        }


class Ed448KeyPair(CryptoKeyPair):
    """
    Represents an Ed448 cryptographic key pair.
    """

    public_bytes: str
    private_bytes: Optional[str]

    @classmethod
    def generate(cls: Type["Ed448KeyPair"]) -> "Ed448KeyPair":
        """
        Generate an Ed448 cryptographic key pair.

        Returns:
            The generated Ed448 key pair.
        """
        new_key = ed448.Ed448PrivateKey.generate()
        return Ed448KeyPair.from_cryptography(new_key)

    @classmethod
    def from_cryptography(
        cls: Type["Ed448KeyPair"],
        key: Union[ed448.Ed448PrivateKey, ed448.Ed448PublicKey],
    ) -> "Ed448KeyPair":
        """
        Create an Ed448KeyPair from a cryptography key.

        Args:
            key: The cryptography key.

        Returns:
            The Ed448 key pair.
        """
        private_bytes = None

        if isinstance(key, ed448.Ed448PrivateKey):
            public_key = key.public_key()
            private_bytes = _byte_to_hex(key.private_bytes_raw())
        elif isinstance(key, ed448.Ed448PublicKey):
            public_key = key
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

        ret = cls(
            public_bytes=_byte_to_hex(public_key.public_bytes_raw()),
            private_bytes=private_bytes,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(
        self,
    ) -> Union[ed448.Ed448PrivateKey, ed448.Ed448PublicKey]:
        if self.private_bytes is not None:
            return ed448.Ed448PrivateKey.from_private_bytes(
                _hex_to_byte(self.private_bytes)
            )
        else:
            return ed448.Ed448PublicKey.from_public_bytes(
                _hex_to_byte(self.public_bytes)
            )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "public_bytes": self.public_bytes,
        }


class KeyPair(InitCryptoParser):
    """
    Represents a cryptographic key pair.

    Attributes:
        algorithm: The key algorithm in string format
        parameters: The dict representation of the key
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    algorithm: str
    parameters: Dict[str, str]

    _init_func = "create"
    _key_pair: CryptoKeyPair

    @classmethod
    def from_cryptography(
        cls: Type["KeyPair"],
        key: Union[
            CertificateIssuerPrivateKeyTypes, CertificateIssuerPublicKeyTypes
        ],
    ) -> "KeyPair":
        """
        Create a KeyPair from a cryptography key.

        Args:
            key: The cryptography key.

        Returns:
            The key pair.
        """
        types = get_args(
            Union[
                CertificateIssuerPrivateKeyTypes,
                CertificateIssuerPublicKeyTypes,
            ]
        )

        name = None
        for check_type in types:
            if isinstance(key, check_type):
                name = check_type.__name__
                break

        class_name = name.replace("PrivateKey", "KeyPair")
        class_name = class_name.replace("PublicKey", "KeyPair")

        key_pair = globals()[class_name].from_cryptography(key)

        ret = cls(
            algorithm=name.replace("Private", ""),
            parameters=key_pair._string_dict(),
            _x509_obj=key_pair._x509_obj,
        )
        ret._x509_obj = key_pair._x509_obj
        ret._key_pair = key_pair
        return ret

    def create(self, key_pair: CryptoKeyPair):
        """
        Create a KeyPair.

        Args:
            key_pair: The cryptographic key pair.
        """
        self._key_pair = key_pair
        self._x509_obj = self._key_pair._to_cryptography()

    def _string_dict(self) -> Dict:
        params = {}
        for k, v in self.parameters.items():
            key = " ".join(ele.title() for ele in k.split("_"))
            if v == "None":
                continue

            params[key] = v

        return {"Public Key Algorithm": self.algorithm, "Parameters": params}

    def _to_cryptography(self) -> CryptoObject:
        return self._crypto_object



from loguru import logger

from pki_tools.crl import _is_revoked
from pki_tools.exceptions import (
    ExtensionMissing,
    OcspInvalidResponseStatus,
    OcspFetchFailure,
    OcspIssuerFetchFailure,
    RevokeCheckFailed,
    Error,
)
from pki_tools.ocsp import _is_revoked_multiple_issuers
from pki_tools.types.certificate import Certificate
from pki_tools.types.chain import Chain
from pki_tools.types.enums import RevokeMode


def is_revoked(
    cert: Certificate,
    chain: Chain,
    crl_cache_seconds: int = 3600,
    ocsp_res_cache_seconds: int = 3600,
    revoke_mode: RevokeMode = RevokeMode.OCSP_FALLBACK_CRL,
) -> bool:
    """
    Checks if a certificate is revoked using OCSP extension and/or
    CRL extension.

    By default, the OCSP is checked first with a fallback to CRL. If you
    only want to check OCSP or only CRL set the "revoke_mode" to either
    [RevokeMode.OCSP_ONLY][pki_tools.types.enums.RevokeMode] or
    [RevokeMode.CRL_ONLY][pki_tools.types.enums.RevokeMode]

    Arguments:
        cert: The [Certificate][pki_tools.types.certificate.Certificate]
            to check revocation for.
        chain: The CA chain including one or more certificates and
            the issuer of the `cert`, signer of the OCSP response and CRL
            issuer. See [Loading Chain][chain] for examples how the
            chain can be created
        crl_cache_seconds: [CRL Only] Specifies how long the CRL should
            be cached, default is 1 hour.
        ocsp_res_cache_seconds: [OCSP Only] Specifies how long the OCSP
            response should be cached, default is 1 hour.
        revoke_mode: A [RevokeMode][pki_tools.types.enums.RevokeMode]
            specifying how to check for revocation, default is OCSP with
            CRL fallback

    Returns:
        True if the certificate is revoked, False otherwise

    Raises:
        SignatureVerificationFailed: When the Chain contains more than one
            certificate and the trust fails either because of some certificate
            has expired or some signature in the chain is invalid
        RevokeCheckFailed: When both OCSP and CRL checks fail
    """
    return is_revoked_multiple_issuers(
        cert,
        chain,
        chain,
        chain,
        crl_cache_seconds=crl_cache_seconds,
        ocsp_res_cache_seconds=ocsp_res_cache_seconds,
        revoke_mode=revoke_mode,
    )


def is_revoked_multiple_issuers(
    cert: Certificate,
    cert_issuer: Chain,
    ocsp_issuer: Chain,
    crl_issuer: Chain,
    crl_cache_seconds: int = 3600,
    ocsp_res_cache_seconds: int = 3600,
    revoke_mode: RevokeMode = RevokeMode.OCSP_FALLBACK_CRL,
) -> bool:
    """
    Checks if a certificate is revoked first using the OCSP extension and then
    the CRL extensions.

    Note that OCSP has precedence over CRL meaning that if OCSP check is
    successful this function will return the bool without checking CRL.

    Otherwise, if OCSP check fails, CRL will be tried next.

    Arguments:
        cert: The [Certificate][pki_tools.types.certificate.Certificate]
            to check revocation for.
        cert_issuer: The CA chain including one or more certificates and
            the issuer of the `cert`. See [Loading Chain][chain]
            for examples how the chain can be created.
        ocsp_issuer: The CA chain including one or more certificates used
            for signing of the OCSP response
        crl_issuer: The CA chain including one or more certificates used
            for signing the CRL
        crl_cache_seconds: [CRL Only] Specifies how long the CRL should
            be cached, default is 1 hour.
        ocsp_res_cache_seconds: [OCSP Only] Specifies how long the OCSP
            response should be cached, default is 1 hour.
        revoke_mode: A [RevokeMode][pki_tools.types.enums.RevokeMode]
            specifying how to check for revocation, default is OCSP with
            CRL fallback

    Returns:
        True if the certificate is revoked, False otherwise

    Raises:
        SignatureVerificationFailed: When the Chain contains more than one
            certificate and the trust fails either because of some certificate
            has expired or some signature in the chain is invalid
        RevokeCheckFailed: When both OCSP and CRL checks fail
    """
    if not revoke_mode == RevokeMode.CRL_ONLY:
        try:
            return _is_revoked_multiple_issuers(
                cert,
                cert_issuer,
                ocsp_issuer,
                ocsp_res_cache_seconds=ocsp_res_cache_seconds,
            )
        except (
            ExtensionMissing,
            OcspInvalidResponseStatus,
            OcspFetchFailure,
            OcspIssuerFetchFailure,
        ) as e:
            if revoke_mode == RevokeMode.OCSP_ONLY:
                err_msg = "OCSP revoke check failed"
                logger.bind(exceptionType=type(e).__name__).error(err_msg)
                raise RevokeCheckFailed(err_msg) from None

            err_msg = "OCSP revoke check failed, trying CRL next"
            logger.bind(exceptionType=type(e).__name__).debug(err_msg)

    try:
        return _is_revoked(cert, crl_issuer, crl_cache_seconds)
    except Error as e:
        err_message = "Revoke checks failed"
        logger.bind(exceptionType=type(e).__name__).error(err_message)
        raise RevokeCheckFailed(err_message) from None



from datetime import datetime

import pytz
from loguru import logger
from pydantic import ConfigDict

from pki_tools.exceptions import (
    NotCompleteChain,
    CertExpired,
    CertIssuerMissingInChain,
)
from pki_tools.types.certificate import Certificate
from pki_tools.types.certificates import Certificates
from pki_tools.types.crl import CertificateRevocationList


class Chain(Certificates):
    """
    Chain holds a list of certificates in a
    [chain of trust](https://en.wikipedia.org/wiki/Chain_of_trust)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def check_chain(self):
        """
        Validate the chain (if it contains more than one certificate)
        checking expiration and signatures of all certificates in the chain

        Raises:
            NotCompleteChain: When the chain contain only one not self-signed
                certificate
            CertExpired: If some certificate in the chain has expired
            InvalidSignedType: When the issuer has a non-supported type
            SignatureVerificationFailed: When the signature verification fails
        """
        if len(self.certificates) == 1:
            if self.certificates[0].issuer == self.certificates[0].subject:
                logger.debug(
                    "Chain contains only one self signed cert, "
                    "nothing to check"
                )
                return
            else:
                raise NotCompleteChain()

        for cert in self.certificates:
            log = logger.bind(subject=cert.subject._string_dict())
            if cert.validity.not_after < datetime.now(
                pytz.utc
            ) or cert.validity.not_before > datetime.now(pytz.utc):
                log.error("Certificate expired")
                raise CertExpired(
                    f"Certificate in chain with subject: '{cert.subject}' "
                    f"has expired"
                )

            issuer = self.get_issuer(cert)
            issuer.verify_signature(cert)

    def get_issuer(
        self,
        signed: [
            Certificate,
            CertificateRevocationList,
        ],
    ) -> Certificate:
        """
        Returns the issuer of a signed entity

        Arguments:
            signed: The signed entity can either be a
                [Certificate][pki_tools.types.certificate.Certificate] or
                [CertificateRevocationList][pki_tools.types.crl.CertificateRevocationList]
        Returns:
            The issuer of the `signed` entity

        Raises:
            CertIssuerMissingInChain: When the issuer of the entity is missing
                in the chain
        """
        cert_issuer = signed.issuer
        log = logger.bind(issuer=cert_issuer._string_dict())

        for next_chain_cert in self.certificates:
            if cert_issuer == next_chain_cert.subject:
                log.trace("Found issuer cert in chain")
                return next_chain_cert

        raise CertIssuerMissingInChain()
